# Copyright 2016:
# * Jim Unroe KC9HI, <rock.unroe@gmail.com>
# * Pavel Milanes CO7WT <pavelmc@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
import logging
import re

from chirp import chirp_common, directory, memmap
from chirp import bitwise, errors, util
from chirp.settings import RadioSettingGroup, RadioSetting, \
    RadioSettingValueBoolean, RadioSettingValueList, \
    RadioSettingValueString, RadioSettingValueInteger, \
    RadioSettingValueFloat, RadioSettings

LOG = logging.getLogger(__name__)

MEM_FORMAT = """
struct mem {
  lbcd rxfreq[4];
  lbcd txfreq[4];
  lbcd rxtone[2];
  lbcd txtone[2];
  u8 unknown0:2,
     txp:2,
     wn:2,
     unknown1:1,
     bcl:1;
  u8 unknown2:2,
     revert:1,
     dname:1,
     unknown3:4;
  u8 unknown4[2];
};

struct nam {
  char name[6];
  u8 unknown1[2];
};

#seekto 0x0000;
struct mem left_memory[500];

#seekto 0x2000;
struct mem right_memory[500];

#seekto 0x4000;
struct nam left_names[500];

#seekto 0x5000;
struct nam right_names[500];

#seekto 0x6000;
u8 left_usedflags[64];

#seekto 0x6040;
u8 left_scanflags[64];

#seekto 0x6080;
u8 right_usedflags[64];

#seekto 0x60C0;
u8 right_scanflags[64];

#seekto 0x6160;
struct {
  char line32[32];
} embedded_msg;

#seekto 0x6180;
struct {
  u8  sbmute:2,        // sub band mute
      unknown1:1,
      workmodb:1,      // work mode (right side)
      dw:1,            // dual watch
      audio:1,         // audio output mode (stereo/mono)
      unknown2:1,
      workmoda:1;      // work mode (left side)
  u8  scansb:1,        // scan stop beep
      aftone:3,        // af tone control
      scand:1,         // scan direction
      scanr:3;         // scan resume
  u8  rxexp:1,         // rx expansion
      ptt:1,           // ptt mode
      display:1,       // display select (frequency/clock)
      omode:1,         // operation mode
      beep:2,          // beep volume
      spkr:2;          // speaker
  u8  cpuclk:1,        // operating mode(cpu clock)
      fkey:3,          // fkey function
      mrscan:1,        // memory scan type
      color:3;         // lcd backlight color
  u8  vox:2,           // vox
      voxs:3,          // vox sensitivity
      mgain:3;         // mic gain
  u8  wbandb:4,        // work band (right side)
      wbanda:4;        // work band (left side)
  u8  sqlb:4,          // squelch level (right side)
      sqla:4;          // squelch level (left side)
  u8  apo:4,           // auto power off
      ars:1,           // automatic repeater shift
      tot:3;           // time out timer
  u8  stepb:4,         // auto step (right side)
      stepa:4;         // auto step (left side)
  u8  rxcoverm:1,      // rx coverage-memory
      lcdc:3,          // lcd contrast
      rxcoverv:1,      // rx coverage-vfo
      lcdb:3;          // lcd brightness
  u8  smode:1,         // smart function mode
      timefmt:1,       // time format
      datefmt:2,       // date format
      timesig:1,       // time signal
      keyb:3;          // key/led brightness
  u8  dwstop:1,        // dual watch stop
      unknown3:1,
      sqlexp:1,        // sql expansion
      decbandsel:1,    // decoding band select
      dtmfmodenc:1,    // dtmf mode encode
      bell:3;          // bell ringer
  u8  unknown4:2,
      btime:6;         // lcd backlight time
  u8  unknown5:2,
      tz:6;            // time zone
  u8  unknown618E;
  u8  unknown618F;
  ul16  offseta;       // work offset (left side)
  ul16  offsetb;       // work offset (right side)
  ul16  mrcha;         // selected memory channel (left)
  ul16  mrchb;         // selected memory channel (right)
  ul16  wpricha;       // work priority channel (left)
  ul16  wprichb;       // work priority channel (right)
  u8  unknown6:3,
      datasql:2,       // data squelch
      dataspd:1,       // data speed
      databnd:2;       // data band select
  u8  unknown7:1,
      pfkey2:3,        // mic p2 key
      unknown8:1,
      pfkey1:3;        // mic p1 key
  u8  unknown9:1,
      pfkey4:3,        // mic p4 key
      unknowna:1,
      pfkey3:3;        // mic p3 key
  u8  unknownb:7,
      dtmfmoddec:1;    // dtmf mode decode
} settings;

#seekto 0x61B0;
struct {
  char line16[16];
} poweron_msg;

#seekto 0x6300;
struct {
  u8  unknown1:3,
      ttdgt:5;         // dtmf digit time
  u8  unknown2:3,
      ttint:5;         // dtmf interval time
  u8  unknown3:3,
      tt1stdgt:5;      // dtmf 1st digit time
  u8  unknown4:3,
      tt1stdly:5;      // dtmf 1st digit delay
  u8  unknown5:3,
      ttdlyqt:5;       // dtmf delay when use qt
  u8  unknown6:3,
      ttdkey:5;        // dtmf d key function
  u8  unknown7;
  u8  unknown8:4,
      ttautod:4;       // dtmf auto dial group
} dtmf;

#seekto 0x6330;
struct {
  u8  unknown1:7,
      ttsig:1;         // dtmf signal
  u8  unknown2:4,
      ttintcode:4;     // dtmf interval code
  u8  unknown3:5,
      ttgrpcode:3;     // dtmf group code
  u8  unknown4:4,
      ttautorst:4;     // dtmf auto reset time
  u8  unknown5:5,
      ttalert:3;       // dtmf alert tone/transpond
} dtmf2;

#seekto 0x6360;
struct {
  u8 code1[8];         // dtmf code
  u8 code1_len;        // dtmf code length
  u8 unknown1[7];
  u8 code2[8];         // dtmf code
  u8 code2_len;        // dtmf code length
  u8 unknown2[7];
  u8 code3[8];         // dtmf code
  u8 code3_len;        // dtmf code length
  u8 unknown3[7];
  u8 code4[8];         // dtmf code
  u8 code4_len;        // dtmf code length
  u8 unknown4[7];
  u8 code5[8];         // dtmf code
  u8 code5_len;        // dtmf code length
  u8 unknown5[7];
  u8 code6[8];         // dtmf code
  u8 code6_len;        // dtmf code length
  u8 unknown6[7];
  u8 code7[8];         // dtmf code
  u8 code7_len;        // dtmf code length
  u8 unknown7[7];
  u8 code8[8];         // dtmf code
  u8 code8_len;        // dtmf code length
  u8 unknown8[7];
  u8 code9[8];         // dtmf code
  u8 code9_len;        // dtmf code length
  u8 unknown9[7];
} dtmfcode;

"""

MEM_SIZE = 0x8000
BLOCK_SIZE = 0x40
MODES = ["FM", "Auto", "NFM", "AM"]
SKIP_VALUES = ["", "S"]
TONES = chirp_common.TONES
DTCS_CODES = chirp_common.DTCS_CODES
NAME_LENGTH = 6
DTMF_CHARS = list("0123456789ABCD*#")
STIMEOUT = 1

# Basic settings lists
LIST_AFTONE = ["Low-3", "Low-2", "Low-1", "Normal", "High-1", "High-2"]
LIST_SPKR = ["Off", "Front", "Rear", "Front + Rear"]
LIST_AUDIO = ["Monaural", "Stereo"]
LIST_SBMUTE = ["Off", "TX", "RX", "Both"]
LIST_MLNHM = ["Min", "Low", "Normal", "High", "Max"]
LIST_PTT = ["Momentary", "Toggle"]
LIST_RXEXP = ["General", "Wide coverage"]
LIST_VOX = ["Off", "Internal mic", "Front hand-mic", "Rear hand-mic"]
LIST_DISPLAY = ["Frequency", "Timer/Clock"]
LIST_MINMAX = ["Min"] + ["%s" % x for x in range(2, 8)] + ["Max"]
LIST_COLOR = ["White-Blue", "Sky-Blue", "Marine-Blue", "Green",
              "Yellow-Green", "Orange", "Amber", "White"]
LIST_BTIME = ["Continuous"] + ["%s" % x for x in range(1, 61)]
LIST_MRSCAN = ["All", "Selected"]
LIST_DWSTOP = ["Auto", "Hold"]
LIST_SCAND = ["Down", "Up"]
LIST_SCANR = ["Busy", "Hold", "1 sec", "3 sec", "5 sec"]
LIST_APO = ["Off", ".5", "1", "1.5"] + ["%s" % x for x in range(2, 13)]
LIST_BEEP = ["Off", "Low", "High"]
LIST_FKEY = ["MHz/AD-F", "AF Dual 1(line-in)", "AF Dual 2(AM)",
             "AF Dual 3(FM)", "PA", "SQL off", "T-call", "WX"]
LIST_PFKEY = ["Off", "SQL off", "TX power", "Scan", "RPT shift", "Reverse",
              "T-Call"]
LIST_AB = ["A", "B"]
LIST_COVERAGE = ["In band", "All"]
LIST_TOT = ["Off"] + ["%s" % x for x in range(5, 25, 5)] + ["30"]
LIST_DATEFMT = ["yyyy/mm/dd", "yyyy/dd/mm", "mm/dd/yyyy", "dd/mm/yyyy"]
LIST_TIMEFMT = ["24H", "12H"]
LIST_TZ = ["-12 INT DL W",
           "-11 MIDWAY",
           "-10 HAST",
           "-9 AKST",
           "-8 PST",
           "-7 MST",
           "-6 CST",
           "-5 EST",
           "-4:30 CARACAS",
           "-4 AST",
           "-3:30 NST",
           "-3 BRASILIA",
           "-2 MATLANTIC",
           "-1 AZORES",
           "-0 LONDON",
           "+0 LONDON",
           "+1 ROME",
           "+2 ATHENS",
           "+3 MOSCOW",
           "+3:30 REHRW",
           "+4 ABUDNABI",
           "+4:30 KABUL",
           "+5 ISLMABAD",
           "+5:30 NEWDELHI",
           "+6 DHAKA",
           "+6:30 YANGON",
           "+7 BANKOK",
           "+8 BEIJING",
           "+9 TOKYO",
           "+10 ADELAIDE",
           "+10 SYDNET",
           "+11 NWCLDNIA",
           "+12 FIJI",
           "+13 NUKALOFA"
           ]
LIST_BELL = ["Off", "1 time", "3 times", "5 times", "8 times", "Continuous"]
LIST_DATABND = ["Main band", "Sub band", "Left band-fixed", "Right band-fixed"]
LIST_DATASPD = ["1200 bps", "9600 bps"]
LIST_DATASQL = ["Busy/TX", "Busy", "TX"]

# Other settings lists
LIST_CPUCLK = ["Clock frequency 1", "Clock frequency 2"]

# Work mode settings lists
LIST_WORK = ["VFO", "Memory System"]
LIST_WBANDB = ["Air", "H-V", "GR1-V", "GR1-U", "H-U", "GR2"]
LIST_WBANDA = ["Line-in", "AM", "FM"] + LIST_WBANDB
LIST_SQL = ["Open"] + ["%s" % x for x in range(1, 10)]
_STEP_LIST = [2.5, 5., 6.25, 8.33, 9., 10., 12.5, 15., 20., 25., 50., 100.,
              200.]
LIST_STEP = ["Auto"] + ["{0:.2f} kHz".format(x) for x in _STEP_LIST]
LIST_SMODE = ["F-1", "F-2"]

# DTMF settings lists
LIST_TTDKEY = ["D code"] + ["Send delay %s s" % x for x in range(1, 17)]
LIST_TT200 = ["%s ms" % x for x in range(50, 210, 10)]
LIST_TT1000 = ["%s ms" % x for x in range(100, 1050, 50)]
LIST_TTSIG = ["Code squelch", "Select call"]
LIST_TTAUTORST = ["Off"] + ["%s s" % x for x in range(1, 16)]
LIST_TTGRPCODE = ["Off"] + list("ABCD*#")
LIST_TTINTCODE = DTMF_CHARS
LIST_TTALERT = ["Off", "Alert tone", "Transpond", "Transpond-ID code",
                "Transpond-transpond code"]
LIST_TTAUTOD = ["%s" % x for x in range(1, 10)]

# valid chars on the LCD
VALID_CHARS = chirp_common.CHARSET_ALPHANUMERIC + \
    "`{|}!\"#$%&'()*+,-./:;<=>?@[]^_"

# Power Levels
POWER_LEVELS = [chirp_common.PowerLevel("Low", watts=5),
                chirp_common.PowerLevel("Mid", watts=20),
                chirp_common.PowerLevel("High", watts=50)]

# B-TECH UV-50X3 id string
UV50X3_id = b"VGC6600MD"


def _clean_buffer(radio):
    radio.pipe.timeout = 0.005
    junk = radio.pipe.read(256)
    radio.pipe.timeout = STIMEOUT
    if junk:
        LOG.debug("Got %i bytes of junk before starting" % len(junk))


def _check_for_double_ack(radio):
    radio.pipe.timeout = 0.005
    c = radio.pipe.read(1)
    radio.pipe.timeout = STIMEOUT
    if c and c != b'\x06':
        _exit_program_mode(radio)
        raise errors.RadioError('Expected nothing or ACK, got %r' % c)


def _rawrecv(radio, amount):
    """Raw read from the radio device"""
    data = b""
    try:
        data = radio.pipe.read(amount)
    except:
        _exit_program_mode(radio)
        msg = "Generic error reading data from radio; check your cable."
        raise errors.RadioError(msg)

    if len(data) != amount:
        _exit_program_mode(radio)
        msg = "Error reading data from radio: not the amount of data we want."
        raise errors.RadioError(msg)

    return data


def _rawsend(radio, data):
    """Raw send to the radio device"""
    try:
        radio.pipe.write(data)
    except:
        raise errors.RadioError("Error sending data to radio")


def _make_frame(cmd, addr, length, data=""):
    """Pack the info in the header format"""
    frame = struct.pack(">BHB", ord(cmd), addr, length)
    # add the data if set
    if len(data) != 0:
        frame += data
    # return the data
    return frame


def _recv(radio, addr, length=BLOCK_SIZE):
    """Get data from the radio """
    # read 4 bytes of header
    hdr = _rawrecv(radio, 4)

    # check for unexpected extra command byte
    c, a, l = struct.unpack(">BHB", hdr)
    if hdr[0:2] == b"WW" and a != addr:
        # extra command byte detected
        # throw away the 1st byte and add the next byte in the buffer
        hdr = hdr[1:] + _rawrecv(radio, 1)

    # read 64 bytes (0x40) of data
    data = _rawrecv(radio, (BLOCK_SIZE))

    # DEBUG
    LOG.info("Response:")
    LOG.debug(util.hexprint(hdr + data))

    c, a, l = struct.unpack(">BHB", hdr)
    if a != addr or l != length or c != ord(b"W"):
        _exit_program_mode(radio)
        LOG.error("Invalid answer for block 0x%04x:" % addr)
        LOG.debug("CMD: %s  ADDR: %04x  SIZE: %02x" % (c, a, l))
        raise errors.RadioError("Unknown response from the radio")

    return data


def _do_ident(radio):
    """Put the radio in PROGRAM mode & identify it"""
    #  set the serial discipline
    radio.pipe.baudrate = 115200
    radio.pipe.parity = "N"
    radio.pipe.timeout = STIMEOUT

    # flush input buffer
    _clean_buffer(radio)

    magic = b"V66LINK"

    _rawsend(radio, magic)

    # Ok, get the ident string
    ident = _rawrecv(radio, 9)

    # check if ident is OK
    if ident != radio.IDENT:
        # bad ident
        msg = "Incorrect model ID, got this:"
        msg += util.hexprint(ident)
        LOG.debug(msg)
        raise errors.RadioError("Radio identification failed.")

    # DEBUG
    LOG.info("Positive ident, got this:")
    LOG.debug(util.hexprint(ident))

    return True


def _exit_program_mode(radio):
    endframe = b"\x45"
    _rawsend(radio, endframe)


def _download(radio):
    """Get the memory map"""

    # put radio in program mode and identify it
    _do_ident(radio)

    # UI progress
    status = chirp_common.Status()
    status.cur = 0
    status.max = MEM_SIZE // BLOCK_SIZE
    status.msg = "Cloning from radio..."
    radio.status_fn(status)

    data = b""
    for addr in range(0, MEM_SIZE, BLOCK_SIZE):
        frame = _make_frame(b"R", addr, BLOCK_SIZE)
        # DEBUG
        LOG.info("Request sent:")
        LOG.debug(util.hexprint(frame))

        # sending the read request
        _rawsend(radio, frame)

        # now we read
        d = _recv(radio, addr)

        # aggregate the data
        data += d

        # UI Update
        status.cur = addr // BLOCK_SIZE
        status.msg = "Cloning from radio..."
        radio.status_fn(status)

    _exit_program_mode(radio)

    return data


def _upload(radio):
    """Upload procedure"""

    MEM_SIZE = 0x7000

    # put radio in program mode and identify it
    _do_ident(radio)

    # UI progress
    status = chirp_common.Status()
    status.cur = 0
    status.max = MEM_SIZE // BLOCK_SIZE
    status.msg = "Cloning to radio..."
    radio.status_fn(status)

    # the fun start here
    for addr in range(0, MEM_SIZE, BLOCK_SIZE):
        # sending the data
        data = radio.get_mmap()[addr:addr + BLOCK_SIZE]

        frame = _make_frame(b"W", addr, BLOCK_SIZE, data)

        _rawsend(radio, frame)

        # receiving the response
        ack = _rawrecv(radio, 1)
        if ack != b"\x06":
            _exit_program_mode(radio)
            msg = "Bad ack writing block 0x%04x" % addr
            raise errors.RadioError(msg)

        _check_for_double_ack(radio)

        # UI Update
        status.cur = addr // BLOCK_SIZE
        status.msg = "Cloning to radio..."
        radio.status_fn(status)

    _exit_program_mode(radio)


def model_match(cls, data):
    """Match the opened/downloaded image to the correct version"""
    rid = data[0x6140:0x6148]

    if rid in cls.IDENT:
        return True

    return False


def _split(rf, f1, f2):
    """Returns False if the two freqs are in the same band (no split)
    or True otherwise"""

    # determine if the two freqs are in the same band
    for low, high in rf.valid_bands:
        if f1 >= low and f1 <= high and \
                f2 >= low and f2 <= high:
            # if the two freqs are on the same Band this is not a split
            return False

    # if you get here is because the freq pairs are split
    return True


class VGCStyleRadio(chirp_common.CloneModeRadio,
                    chirp_common.ExperimentalRadio):
    """BTECH's UV-50X3"""
    VENDOR = "BTECH"
    NEEDS_COMPAT_SERIAL = False
    _air_range = (108000000, 136000000)
    _vhf_range = (136000000, 174000000)
    _vhf2_range = (174000000, 250000000)
    _220_range = (222000000, 225000000)
    _gen1_range = (300000000, 400000000)
    _uhf_range = (400000000, 480000000)
    _gen2_range = (480000000, 520000000)
    _upper = 499
    MODEL = ""
    IDENT = ""

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = \
            ('The UV-50X3 driver is a beta version.\n'
             '\n'
             'Please save an unedited copy of your first successful\n'
             'download to a CHIRP Radio Images(*.img) file.'
             )
        rp.pre_download = _(
            "Follow this instructions to download your info:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable\n"
            "3 - Turn on your radio\n"
            "4 - Do the download of your radio data\n")
        rp.pre_upload = _(
            "Follow this instructions to upload your info:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable\n"
            "3 - Turn on your radio\n"
            "4 - Do the upload of your radio data\n")
        return rp

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_bank = False
        rf.has_tuning_step = False
        rf.can_odd_split = True
        rf.has_name = True
        rf.has_offset = True
        rf.has_mode = True
        rf.has_dtcs = True
        rf.has_rx_dtcs = True
        rf.has_dtcs_polarity = True
        rf.has_ctone = True
        rf.has_cross = True
        rf.has_sub_devices = self.VARIANT == ""
        rf.valid_modes = MODES
        rf.valid_characters = VALID_CHARS
        rf.valid_duplexes = ["", "-", "+", "split", "off"]
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'DTCS', 'Cross']
        rf.valid_cross_modes = [
            "Tone->Tone",
            "DTCS->",
            "->DTCS",
            "Tone->DTCS",
            "DTCS->Tone",
            "->Tone",
            "DTCS->DTCS"]
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_skips = SKIP_VALUES
        rf.valid_name_length = NAME_LENGTH
        rf.valid_dtcs_codes = DTCS_CODES
        rf.valid_tuning_steps = _STEP_LIST
        rf.valid_bands = [self._air_range,
                          self._vhf_range,
                          self._vhf2_range,
                          self._220_range,
                          self._gen1_range,
                          self._uhf_range,
                          self._gen2_range]
        rf.memory_bounds = (0, self._upper)
        return rf

    def get_sub_devices(self):
        return [UV50X3Left(self._mmap), UV50X3Right(self._mmap)]

    def sync_in(self):
        """Download from radio"""
        try:
            data = _download(self)
        except errors.RadioError:
            # Pass through any real errors we raise
            raise
        except:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during download')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')
        self._mmap = memmap.MemoryMapBytes(data)
        self.process_mmap()

    def sync_out(self):
        """Upload to radio"""
        try:
            _upload(self)
        except:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during upload')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')

    def process_mmap(self):
        """Process the mem map into the mem object"""
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number])

    def decode_tone(self, val):
        """Parse the tone data to decode from mem, it returns:
        Mode (''|DTCS|Tone), Value (None|###), Polarity (None,N,R)"""
        if val.get_raw(asbytes=False) == "\xFF\xFF":
            return '', None, None

        val = int(val)
        if val >= 12000:
            a = val - 12000
            return 'DTCS', a, 'R'
        elif val >= 8000:
            a = val - 8000
            return 'DTCS', a, 'N'
        else:
            a = val / 10.0
            return 'Tone', a, None

    def encode_tone(self, memval, mode, value, pol):
        """Parse the tone data to encode from UI to mem"""
        if mode == '':
            memval[0].set_raw(0xFF)
            memval[1].set_raw(0xFF)
        elif mode == 'Tone':
            memval.set_value(int(value * 10))
        elif mode == 'DTCS':
            flag = 0x80 if pol == 'N' else 0xC0
            memval.set_value(value)
            memval[1].set_bits(flag)
        else:
            raise Exception("Internal error: invalid mode `%s'" % mode)

    def _memory_obj(self, suffix=""):
        return getattr(self._memobj, "%s_memory%s" % (self._vfo, suffix))

    def _name_obj(self, suffix=""):
        return getattr(self._memobj, "%s_names%s" % (self._vfo, suffix))

    def _scan_obj(self, suffix=""):
        return getattr(self._memobj, "%s_scanflags%s" % (self._vfo, suffix))

    def _used_obj(self, suffix=""):
        return getattr(self._memobj, "%s_usedflags%s" % (self._vfo, suffix))

    def get_memory(self, number):
        """Get the mem representation from the radio image"""
        bitpos = (1 << (number % 8))
        bytepos = (number / 8)

        _mem = self._memory_obj()[number]
        _names = self._name_obj()[number]
        _scn = self._scan_obj()[bytepos]
        _usd = self._used_obj()[bytepos]

        isused = bitpos & int(_usd)
        isscan = bitpos & int(_scn)

        # Create a high-level memory object to return to the UI
        mem = chirp_common.Memory()

        # Memory number
        mem.number = number

        if not isused:
            mem.empty = True
            return mem

        # Freq and offset
        mem.freq = int(_mem.rxfreq) * 10
        # tx freq can be blank
        if _mem.get_raw(asbytes=False)[4] == "\xFF":
            # TX freq not set
            mem.offset = 0
            mem.duplex = "off"
        else:
            # TX feq set
            offset = (int(_mem.txfreq) * 10) - mem.freq
            if offset != 0:
                if _split(self.get_features(), mem.freq, int(
                          _mem.txfreq) * 10):
                    mem.duplex = "split"
                    mem.offset = int(_mem.txfreq) * 10
                elif offset < 0:
                    mem.offset = abs(offset)
                    mem.duplex = "-"
                elif offset > 0:
                    mem.offset = offset
                    mem.duplex = "+"
            else:
                mem.offset = 0

        # skip
        if not isscan:
            mem.skip = "S"

        # name TAG of the channel
        mem.name = str(_names.name).strip("\xFF")

        # power
        mem.power = POWER_LEVELS[int(_mem.txp)]

        # wide/narrow
        mem.mode = MODES[int(_mem.wn)]

        # tone data
        rxtone = txtone = None
        txtone = self.decode_tone(_mem.txtone)
        rxtone = self.decode_tone(_mem.rxtone)
        chirp_common.split_tone_decode(mem, txtone, rxtone)

        # Extra
        mem.extra = RadioSettingGroup("extra", "Extra")

        bcl = RadioSetting("bcl", "Busy channel lockout",
                           RadioSettingValueBoolean(bool(_mem.bcl)))
        mem.extra.append(bcl)

        revert = RadioSetting("revert", "Revert",
                              RadioSettingValueBoolean(bool(_mem.revert)))
        mem.extra.append(revert)

        dname = RadioSetting("dname", "Display name",
                             RadioSettingValueBoolean(bool(_mem.dname)))
        mem.extra.append(dname)

        return mem

    def set_memory(self, mem):
        """Set the memory data in the eeprom img from the UI"""
        bitpos = (1 << (mem.number % 8))
        bytepos = (mem.number / 8)

        _mem = self._memory_obj()[mem.number]
        _names = self._name_obj()[mem.number]
        _scn = self._scan_obj()[bytepos]
        _usd = self._used_obj()[bytepos]

        if mem.empty:
            _usd &= ~bitpos
            _scn &= ~bitpos
            _mem.set_raw("\xFF" * 16)
            _names.name = ("\xFF" * 6)
            return
        else:
            _usd |= bitpos

        # frequency
        _mem.rxfreq = mem.freq / 10

        # duplex
        if mem.duplex == "+":
            _mem.txfreq = (mem.freq + mem.offset) / 10
        elif mem.duplex == "-":
            _mem.txfreq = (mem.freq - mem.offset) / 10
        elif mem.duplex == "off":
            for i in _mem.txfreq:
                i.set_raw("\xFF")
        elif mem.duplex == "split":
            _mem.txfreq = mem.offset / 10
        else:
            _mem.txfreq = mem.freq / 10

        # tone data
        ((txmode, txtone, txpol), (rxmode, rxtone, rxpol)) = \
            chirp_common.split_tone_encode(mem)
        self.encode_tone(_mem.txtone, txmode, txtone, txpol)
        self.encode_tone(_mem.rxtone, rxmode, rxtone, rxpol)

        # name TAG of the channel
        _names.name = mem.name.rstrip().ljust(6, "\xFF")

        # power level, # default power level is low
        _mem.txp = 0 if mem.power is None else POWER_LEVELS.index(mem.power)

        # wide/narrow
        _mem.wn = MODES.index(mem.mode)

        if mem.skip == "S":
            _scn &= ~bitpos
        else:
            _scn |= bitpos

        # autoset display to display name if filled
        if mem.extra:
            # mem.extra only seems to be populated when called from edit panel
            dname = mem.extra["dname"]
        else:
            dname = None
        if mem.name:
            _mem.dname = True
            if dname and not dname.changed():
                dname.value = True
        else:
            _mem.dname = False
            if dname and not dname.changed():
                dname.value = False

        # resetting unknowns, this has to be set by hand
        _mem.unknown0 = 0
        _mem.unknown1 = 0
        _mem.unknown2 = 0
        _mem.unknown3 = 0

        # extra settings
        if len(mem.extra) > 0:
            # there are setting, parse
            for setting in mem.extra:
                setattr(_mem, setting.get_name(), setting.value)
        else:
            # there are no extra settings, load defaults
            _mem.bcl = 0
            _mem.revert = 0
            _mem.dname = 1

    def _bbcd2dtmf(self, bcdarr, strlen=16):
        # doing bbcd, but with support for ABCD*#
        LOG.debug(bcdarr.get_value())
        string = ''.join("%02X" % b for b in bcdarr)
        LOG.debug("@_bbcd2dtmf, received: %s" % string)
        string = string.replace('E', '*').replace('F', '#')
        if strlen <= 16:
            string = string[:strlen]
        return string

    def _dtmf2bbcd(self, value):
        dtmfstr = value.get_value()
        dtmfstr = dtmfstr.replace('*', 'E').replace('#', 'F')
        dtmfstr = str.ljust(dtmfstr.strip(), 16, "F")
        bcdarr = list(bytearray.fromhex(dtmfstr))
        LOG.debug("@_dtmf2bbcd, sending: %s" % bcdarr)
        return bcdarr

    def get_settings(self):
        """Translate the bit in the mem_struct into settings in the UI"""
        _mem = self._memobj
        basic = RadioSettingGroup("basic", "Basic Settings")
        other = RadioSettingGroup("other", "Other Settings")
        work = RadioSettingGroup("work", "Work Mode Settings")
        dtmf = RadioSettingGroup("dtmf", "DTMF Settings")
        top = RadioSettings(basic, other, work, dtmf)

        # Basic

        # Audio: A01-A04

        aftone = RadioSetting("settings.aftone", "AF tone control",
                              RadioSettingValueList(LIST_AFTONE, LIST_AFTONE[
                                  _mem.settings.aftone]))
        basic.append(aftone)

        spkr = RadioSetting("settings.spkr", "Speaker",
                            RadioSettingValueList(LIST_SPKR, LIST_SPKR[
                                _mem.settings.spkr]))
        basic.append(spkr)

        audio = RadioSetting("settings.audio", "Stereo/Mono",
                             RadioSettingValueList(LIST_AUDIO, LIST_AUDIO[
                                 _mem.settings.audio]))
        basic.append(audio)

        sbmute = RadioSetting("settings.sbmute", "Sub band mute",
                              RadioSettingValueList(LIST_SBMUTE, LIST_SBMUTE[
                                  _mem.settings.sbmute]))
        basic.append(sbmute)

        # TX/RX: B01-B08

        mgain = RadioSetting("settings.mgain", "Mic gain",
                             RadioSettingValueList(LIST_MLNHM, LIST_MLNHM[
                                 _mem.settings.mgain]))
        basic.append(mgain)

        ptt = RadioSetting("settings.ptt", "PTT mode",
                           RadioSettingValueList(LIST_PTT, LIST_PTT[
                               _mem.settings.ptt]))
        basic.append(ptt)

        # B03 (per channel)
        # B04 (per channel)

        rxexp = RadioSetting("settings.rxexp", "RX expansion",
                             RadioSettingValueList(LIST_RXEXP, LIST_RXEXP[
                                 _mem.settings.rxexp]))
        basic.append(rxexp)

        vox = RadioSetting("settings.vox", "Vox",
                           RadioSettingValueList(LIST_VOX, LIST_VOX[
                               _mem.settings.vox]))
        basic.append(vox)

        voxs = RadioSetting("settings.voxs", "Vox sensitivity",
                            RadioSettingValueList(LIST_MLNHM, LIST_MLNHM[
                                _mem.settings.voxs]))
        basic.append(voxs)

        # B08 (per channel)

        # Display: C01-C06

        display = RadioSetting("settings.display", "Display select",
                               RadioSettingValueList(
                                   LIST_DISPLAY, LIST_DISPLAY[
                                       _mem.settings.display]))
        basic.append(display)

        lcdb = RadioSetting("settings.lcdb", "LCD brightness",
                            RadioSettingValueList(LIST_MINMAX, LIST_MINMAX[
                                _mem.settings.lcdb]))
        basic.append(lcdb)

        color = RadioSetting("settings.color", "LCD color",
                             RadioSettingValueList(LIST_COLOR, LIST_COLOR[
                                 _mem.settings.color]))
        basic.append(color)

        lcdc = RadioSetting("settings.lcdc", "LCD contrast",
                            RadioSettingValueList(LIST_MINMAX, LIST_MINMAX[
                                _mem.settings.lcdc]))
        basic.append(lcdc)

        btime = RadioSetting("settings.btime", "LCD backlight time",
                             RadioSettingValueList(LIST_BTIME, LIST_BTIME[
                                 _mem.settings.btime]))
        basic.append(btime)

        keyb = RadioSetting("settings.keyb", "Key brightness",
                            RadioSettingValueList(LIST_MINMAX, LIST_MINMAX[
                                _mem.settings.keyb]))
        basic.append(keyb)

        # Memory: D01-D04

        # D01 (per channel)
        # D02 (per channel)

        mrscan = RadioSetting("settings.mrscan", "Memory scan type",
                              RadioSettingValueList(LIST_MRSCAN, LIST_MRSCAN[
                                  _mem.settings.mrscan]))
        basic.append(mrscan)

        # D04 (per channel)

        # Scan: E01-E04

        dwstop = RadioSetting("settings.dwstop", "Dual watch stop",
                              RadioSettingValueList(LIST_DWSTOP, LIST_DWSTOP[
                                  _mem.settings.dwstop]))
        basic.append(dwstop)

        scand = RadioSetting("settings.scand", "Scan direction",
                             RadioSettingValueList(LIST_SCAND, LIST_SCAND[
                                 _mem.settings.scand]))
        basic.append(scand)

        scanr = RadioSetting("settings.scanr", "Scan resume",
                             RadioSettingValueList(LIST_SCANR, LIST_SCANR[
                                 _mem.settings.scanr]))
        basic.append(scanr)

        scansb = RadioSetting("settings.scansb", "Scan stop beep",
                              RadioSettingValueBoolean(_mem.settings.scansb))
        basic.append(scansb)

        # System: F01-F09

        apo = RadioSetting("settings.apo", "Automatic power off [hours]",
                           RadioSettingValueList(LIST_APO, LIST_APO[
                               _mem.settings.apo]))
        basic.append(apo)

        ars = RadioSetting("settings.ars", "Automatic repeater shift",
                           RadioSettingValueBoolean(_mem.settings.ars))
        basic.append(ars)

        beep = RadioSetting("settings.beep", "Beep volume",
                            RadioSettingValueList(LIST_BEEP, LIST_BEEP[
                                _mem.settings.beep]))
        basic.append(beep)

        fkey = RadioSetting("settings.fkey", "F key",
                            RadioSettingValueList(LIST_FKEY, LIST_FKEY[
                                _mem.settings.fkey]))
        basic.append(fkey)

        pfkey1 = RadioSetting("settings.pfkey1", "Mic P1 key",
                              RadioSettingValueList(LIST_PFKEY, LIST_PFKEY[
                                  _mem.settings.pfkey1]))
        basic.append(pfkey1)

        pfkey2 = RadioSetting("settings.pfkey2", "Mic P2 key",
                              RadioSettingValueList(LIST_PFKEY, LIST_PFKEY[
                                  _mem.settings.pfkey2]))
        basic.append(pfkey2)

        pfkey3 = RadioSetting("settings.pfkey3", "Mic P3 key",
                              RadioSettingValueList(LIST_PFKEY, LIST_PFKEY[
                                  _mem.settings.pfkey3]))
        basic.append(pfkey3)

        pfkey4 = RadioSetting("settings.pfkey4", "Mic P4 key",
                              RadioSettingValueList(LIST_PFKEY, LIST_PFKEY[
                                  _mem.settings.pfkey4]))
        basic.append(pfkey4)

        omode = RadioSetting("settings.omode", "Operation mode",
                             RadioSettingValueList(LIST_AB, LIST_AB[
                                 _mem.settings.omode]))
        basic.append(omode)

        rxcoverm = RadioSetting("settings.rxcoverm", "RX coverage - memory",
                                RadioSettingValueList(
                                    LIST_COVERAGE, LIST_COVERAGE[
                                        _mem.settings.rxcoverm]))
        basic.append(rxcoverm)

        rxcoverv = RadioSetting("settings.rxcoverv", "RX coverage - VFO",
                                RadioSettingValueList(
                                    LIST_COVERAGE, LIST_COVERAGE[
                                        _mem.settings.rxcoverv]))
        basic.append(rxcoverv)

        tot = RadioSetting("settings.tot", "Time out timer [min]",
                           RadioSettingValueList(LIST_TOT, LIST_TOT[
                                                 _mem.settings.tot]))
        basic.append(tot)

        # Timer/Clock: G01-G04

        # G01
        datefmt = RadioSetting("settings.datefmt", "Date format",
                               RadioSettingValueList(
                                   LIST_DATEFMT, LIST_DATEFMT[
                                       _mem.settings.datefmt]))
        basic.append(datefmt)

        timefmt = RadioSetting("settings.timefmt", "Time format",
                               RadioSettingValueList(
                                   LIST_TIMEFMT, LIST_TIMEFMT[
                                       _mem.settings.timefmt]))
        basic.append(timefmt)

        timesig = RadioSetting("settings.timesig", "Time signal",
                               RadioSettingValueBoolean(_mem.settings.timesig))
        basic.append(timesig)

        tz = RadioSetting("settings.tz", "Time zone",
                          RadioSettingValueList(LIST_TZ, LIST_TZ[
                              _mem.settings.tz]))
        basic.append(tz)

        # Signaling: H01-H06

        bell = RadioSetting("settings.bell", "Bell ringer",
                            RadioSettingValueList(LIST_BELL, LIST_BELL[
                                _mem.settings.bell]))
        basic.append(bell)

        # H02 (per channel)

        dtmfmodenc = RadioSetting("settings.dtmfmodenc", "DTMF mode encode",
                                  RadioSettingValueBoolean(
                                      _mem.settings.dtmfmodenc))
        basic.append(dtmfmodenc)

        dtmfmoddec = RadioSetting("settings.dtmfmoddec", "DTMF mode decode",
                                  RadioSettingValueBoolean(
                                      _mem.settings.dtmfmoddec))
        basic.append(dtmfmoddec)

        # H04 (per channel)

        decbandsel = RadioSetting("settings.decbandsel", "DTMF band select",
                                  RadioSettingValueList(LIST_AB, LIST_AB[
                                      _mem.settings.decbandsel]))
        basic.append(decbandsel)

        sqlexp = RadioSetting("settings.sqlexp", "SQL expansion",
                              RadioSettingValueBoolean(_mem.settings.sqlexp))
        basic.append(sqlexp)

        # Pkt: I01-I03

        databnd = RadioSetting("settings.databnd", "Packet data band",
                               RadioSettingValueList(
                                   LIST_DATABND, LIST_DATABND[
                                       _mem.settings.databnd]))
        basic.append(databnd)

        dataspd = RadioSetting("settings.dataspd", "Packet data speed",
                               RadioSettingValueList(
                                   LIST_DATASPD, LIST_DATASPD[
                                       _mem.settings.dataspd]))
        basic.append(dataspd)

        datasql = RadioSetting("settings.datasql", "Packet data squelch",
                               RadioSettingValueList(
                                   LIST_DATASQL, LIST_DATASQL[
                                       _mem.settings.datasql]))
        basic.append(datasql)

        # Other

        dw = RadioSetting("settings.dw", "Dual watch",
                          RadioSettingValueBoolean(_mem.settings.dw))
        other.append(dw)

        cpuclk = RadioSetting("settings.cpuclk", "CPU clock frequency",
                              RadioSettingValueList(LIST_CPUCLK, LIST_CPUCLK[
                                  _mem.settings.cpuclk]))
        other.append(cpuclk)

        def _filter(name):
            filtered = ""
            for char in str(name):
                if char in VALID_CHARS:
                    filtered += char
                else:
                    filtered += " "
            return filtered

        line16 = RadioSetting("poweron_msg.line16", "Power-on message",
                              RadioSettingValueString(0, 16, _filter(
                                  _mem.poweron_msg.line16)))
        other.append(line16)

        line32 = RadioSetting("embedded_msg.line32", "Embedded message",
                              RadioSettingValueString(0, 32, _filter(
                                  _mem.embedded_msg.line32)))
        other.append(line32)

        # Work

        workmoda = RadioSetting("settings.workmoda", "Work mode A",
                                RadioSettingValueList(LIST_WORK, LIST_WORK[
                                    _mem.settings.workmoda]))
        work.append(workmoda)

        workmodb = RadioSetting("settings.workmodb", "Work mode B",
                                RadioSettingValueList(LIST_WORK, LIST_WORK[
                                    _mem.settings.workmodb]))
        work.append(workmodb)

        wbanda = RadioSetting("settings.wbanda", "Work band A",
                              RadioSettingValueList(LIST_WBANDA, LIST_WBANDA[
                                  (_mem.settings.wbanda) - 1]))
        work.append(wbanda)

        wbandb = RadioSetting("settings.wbandb", "Work band B",
                              RadioSettingValueList(LIST_WBANDB, LIST_WBANDB[
                                  (_mem.settings.wbandb) - 4]))
        work.append(wbandb)

        sqla = RadioSetting("settings.sqla", "Squelch A",
                            RadioSettingValueList(LIST_SQL, LIST_SQL[
                                _mem.settings.sqla]))
        work.append(sqla)

        sqlb = RadioSetting("settings.sqlb", "Squelch B",
                            RadioSettingValueList(LIST_SQL, LIST_SQL[
                                _mem.settings.sqlb]))
        work.append(sqlb)

        stepa = RadioSetting("settings.stepa", "Auto step A",
                             RadioSettingValueList(LIST_STEP, LIST_STEP[
                                 _mem.settings.stepa]))
        work.append(stepa)

        stepb = RadioSetting("settings.stepb", "Auto step B",
                             RadioSettingValueList(LIST_STEP, LIST_STEP[
                                 _mem.settings.stepb]))
        work.append(stepb)

        mrcha = RadioSetting("settings.mrcha", "Current channel A",
                             RadioSettingValueInteger(0, 499,
                                                      _mem.settings.mrcha))
        work.append(mrcha)

        mrchb = RadioSetting("settings.mrchb", "Current channel B",
                             RadioSettingValueInteger(0, 499,
                                                      _mem.settings.mrchb))
        work.append(mrchb)

        val = _mem.settings.offseta / 100.00
        offseta = RadioSetting("settings.offseta", "Offset A (0-37.95)",
                               RadioSettingValueFloat(0, 38.00, val, 0.05, 2))
        work.append(offseta)

        val = _mem.settings.offsetb / 100.00
        offsetb = RadioSetting("settings.offsetb", "Offset B (0-79.95)",
                               RadioSettingValueFloat(0, 80.00, val, 0.05, 2))
        work.append(offsetb)

        wpricha = RadioSetting("settings.wpricha", "Priority channel A",
                               RadioSettingValueInteger(0, 499,
                                                        _mem.settings.wpricha))
        work.append(wpricha)

        wprichb = RadioSetting("settings.wprichb", "Priority channel B",
                               RadioSettingValueInteger(0, 499,
                                                        _mem.settings.wprichb))
        work.append(wprichb)

        smode = RadioSetting("settings.smode", "Smart function mode",
                             RadioSettingValueList(LIST_SMODE, LIST_SMODE[
                                 _mem.settings.smode]))
        work.append(smode)

        # dtmf

        ttdkey = RadioSetting("dtmf.ttdkey", "D key function",
                              RadioSettingValueList(LIST_TTDKEY, LIST_TTDKEY[
                                  _mem.dtmf.ttdkey]))
        dtmf.append(ttdkey)

        ttdgt = RadioSetting("dtmf.ttdgt", "Digit time",
                             RadioSettingValueList(LIST_TT200, LIST_TT200[
                                 (_mem.dtmf.ttdgt) - 5]))
        dtmf.append(ttdgt)

        ttint = RadioSetting("dtmf.ttint", "Interval time",
                             RadioSettingValueList(LIST_TT200, LIST_TT200[
                                 (_mem.dtmf.ttint) - 5]))
        dtmf.append(ttint)

        tt1stdgt = RadioSetting("dtmf.tt1stdgt", "1st digit time",
                                RadioSettingValueList(LIST_TT200, LIST_TT200[
                                    (_mem.dtmf.tt1stdgt) - 5]))
        dtmf.append(tt1stdgt)

        tt1stdly = RadioSetting("dtmf.tt1stdly", "1st digit delay time",
                                RadioSettingValueList(LIST_TT1000, LIST_TT1000[
                                    (_mem.dtmf.tt1stdly) - 2]))
        dtmf.append(tt1stdly)

        ttdlyqt = RadioSetting("dtmf.ttdlyqt", "Digit delay when use qt",
                               RadioSettingValueList(LIST_TT1000, LIST_TT1000[
                                   (_mem.dtmf.ttdlyqt) - 2]))
        dtmf.append(ttdlyqt)

        ttsig = RadioSetting("dtmf2.ttsig", "Signal",
                             RadioSettingValueList(LIST_TTSIG, LIST_TTSIG[
                                 _mem.dtmf2.ttsig]))
        dtmf.append(ttsig)

        ttautorst = RadioSetting("dtmf2.ttautorst", "Auto reset time",
                                 RadioSettingValueList(
                                     LIST_TTAUTORST, LIST_TTAUTORST[
                                         _mem.dtmf2.ttautorst]))
        dtmf.append(ttautorst)

        if _mem.dtmf2.ttgrpcode > 0x06:
            val = 0x00
        else:
            val = _mem.dtmf2.ttgrpcode
        ttgrpcode = RadioSetting("dtmf2.ttgrpcode", "Group code",
                                 RadioSettingValueList(LIST_TTGRPCODE,
                                                       LIST_TTGRPCODE[val]))
        dtmf.append(ttgrpcode)

        ttintcode = RadioSetting("dtmf2.ttintcode", "Interval code",
                                 RadioSettingValueList(
                                     LIST_TTINTCODE, LIST_TTINTCODE[
                                         _mem.dtmf2.ttintcode]))
        dtmf.append(ttintcode)

        if _mem.dtmf2.ttalert > 0x04:
            val = 0x00
        else:
            val = _mem.dtmf2.ttalert
        ttalert = RadioSetting("dtmf2.ttalert", "Alert tone/transpond",
                               RadioSettingValueList(LIST_TTALERT,
                                                     LIST_TTALERT[val]))
        dtmf.append(ttalert)

        ttautod = RadioSetting("dtmf.ttautod", "Auto dial group",
                               RadioSettingValueList(
                                   LIST_TTAUTOD, LIST_TTAUTOD[
                                       _mem.dtmf.ttautod]))
        dtmf.append(ttautod)

        # setup 9 dtmf autodial entries
        for i in map(str, list(range(1, 10))):
            objname = "code" + i
            strname = "Code " + str(i)
            dtmfsetting = getattr(_mem.dtmfcode, objname)
            dtmflen = getattr(_mem.dtmfcode, objname + "_len")
            dtmfstr = self._bbcd2dtmf(dtmfsetting, dtmflen)
            code = RadioSettingValueString(0, 16, dtmfstr)
            code.set_charset(DTMF_CHARS + list(" "))
            rs = RadioSetting("dtmfcode." + objname, strname, code)
            dtmf.append(rs)
        return top

    def set_settings(self, settings):
        _settings = self._memobj.settings
        _mem = self._memobj
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    name = element.get_name()
                    if "." in name:
                        bits = name.split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            if "/" in bit:
                                bit, index = bit.split("/", 1)
                                index = int(index)
                                obj = getattr(obj, bit)[index]
                            else:
                                obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = _settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("Using apply callback")
                        element.run_apply_callback()
                    elif setting == "line16":
                        setattr(obj, setting, str(element.value).rstrip(
                            " ").ljust(16, "\xFF"))
                    elif setting == "line32":
                        setattr(obj, setting, str(element.value).rstrip(
                            " ").ljust(32, "\xFF"))
                    elif setting == "wbanda":
                        setattr(obj, setting, int(element.value) + 1)
                    elif setting == "wbandb":
                        setattr(obj, setting, int(element.value) + 4)
                    elif setting in ["offseta", "offsetb"]:
                        val = element.value
                        value = int(val.get_value() * 100)
                        setattr(obj, setting, value)
                    elif setting in ["ttdgt", "ttint", "tt1stdgt"]:
                        setattr(obj, setting, int(element.value) + 5)
                    elif setting in ["tt1stdly", "ttdlyqt"]:
                        setattr(obj, setting, int(element.value) + 2)
                    elif re.match(r'code\d', setting):
                        # set dtmf length field and then get bcd dtmf
                        dtmfstrlen = len(str(element.value).strip())
                        setattr(_mem.dtmfcode, setting + "_len", dtmfstrlen)
                        dtmfstr = self._dtmf2bbcd(element.value)
                        setattr(_mem.dtmfcode, setting, dtmfstr)
                    elif element.value.get_mutable():
                        LOG.debug("Setting %s = %s" % (setting, element.value))
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise

    @classmethod
    def match_model(cls, filedata, filename):
        match_size = False
        match_model = False

        # testing the file data size
        if len(filedata) == MEM_SIZE:
            match_size = True

        # testing the firmware model fingerprint
        match_model = model_match(cls, filedata)

        if match_size and match_model:
            return True
        else:
            return False


@directory.register
class UV50X3(VGCStyleRadio):
    """BTech UV-50X3"""
    MODEL = "UV-50X3"
    IDENT = UV50X3_id


class UV50X3Left(UV50X3):
    VARIANT = "Left"
    _vfo = "left"


class UV50X3Right(UV50X3):
    VARIANT = "Right"
    _vfo = "right"



# Copyright 2008 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import csv
import logging

from chirp import chirp_common, errors, directory

LOG = logging.getLogger(__name__)
DEFAULT_POWER_LEVEL = chirp_common.AutoNamedPowerLevel(50)


class OmittedHeaderError(Exception):
    """Internal exception to signal that a column has been omitted"""
    pass


def get_datum_by_header(headers, data, header):
    """Return the column corresponding to @headers[@header] from @data"""
    if header not in headers:
        raise OmittedHeaderError("Header %s not provided" % header)

    try:
        return data[headers.index(header)]
    except IndexError:
        raise OmittedHeaderError("Header %s not provided on this line" %
                                 header)


def write_memory(writer, mem):
    """Write @mem using @writer if not empty"""
    if mem.empty:
        return
    writer.writerow(mem.to_csv())


def parse_cross_mode(value):
    if value not in chirp_common.CROSS_MODES:
        raise ValueError('Invalid cross mode %r' % value)
    return value


@directory.register
class CSVRadio(chirp_common.FileBackedRadio):
    """A driver for Generic CSV files"""
    VENDOR = "Generic"
    MODEL = "CSV"
    FILE_EXTENSION = "csv"
    FORMATS = [directory.register_format('CSV', '*.csv')]

    ATTR_MAP = {
        "Location":      (int,   "number"),
        "Name":          (str,   "name"),
        "Frequency":     (chirp_common.parse_freq, "freq"),
        "Duplex":        (str,   "duplex"),
        "Offset":        (chirp_common.parse_freq, "offset"),
        "Tone":          (str,   "tmode"),
        "rToneFreq":     (float, "rtone"),
        "cToneFreq":     (float, "ctone"),
        "DtcsCode":      (int,   "dtcs"),
        "DtcsPolarity":  (str,   "dtcs_polarity"),
        "RxDtcsCode":    (int,   "rx_dtcs"),
        "CrossMode":     (parse_cross_mode, "cross_mode"),
        "Mode":          (str,   "mode"),
        "TStep":         (float, "tuning_step"),
        "Skip":          (str,   "skip"),
        "Power":         (chirp_common.parse_power, "power"),
        "Comment":       (str,   "comment"),
        }

    def _blank(self, setDefault=False):
        self.errors = []
        self.memories = [chirp_common.Memory(i, True) for i in range(0, 1000)]
        if (setDefault):
            self.memories[0].empty = False
            self.memories[0].freq = 146010000
            # Default to 50W
            self.memories[0].power = DEFAULT_POWER_LEVEL

    def __init__(self, pipe):
        chirp_common.FileBackedRadio.__init__(self, None)
        self.memories = []
        self.file_has_rTone = None  # Set in load(), used in _clean_tmode()
        self.file_has_cTone = None

        # Persistence for comment lines
        # List of tuples of (previous_memory, comment)
        self._comments = []

        self._filename = pipe
        if self._filename and os.path.exists(self._filename):
            self.load()
        else:
            self._blank(True)

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_bank = False
        rf.requires_call_lists = False
        rf.has_implicit_calls = False
        rf.memory_bounds = (0, len(self.memories)-1)
        rf.has_infinite_number = True
        rf.has_nostep_tuning = True
        rf.has_comment = True
        rf.has_rx_dtcs = True
        rf.has_variable_power = True
        rf.can_odd_split = True

        rf.valid_modes = list(chirp_common.MODES)
        rf.valid_tmodes = list(chirp_common.TONE_MODES)
        rf.valid_cross_modes = list(chirp_common.CROSS_MODES)
        rf.valid_duplexes = ["", "-", "+", "split", "off"]
        rf.valid_tuning_steps = list(chirp_common.TUNING_STEPS)
        rf.valid_bands = [(1, 10000000000)]
        rf.valid_skips = ["", "S"]
        rf.valid_characters = chirp_common.CHARSET_1252
        rf.valid_name_length = 999
        rf.valid_power_levels = [chirp_common.AutoNamedPowerLevel(0.1),
                                 DEFAULT_POWER_LEVEL,
                                 chirp_common.AutoNamedPowerLevel(1500)]

        return rf

    def _clean(self, headers, line, mem):
        """Runs post-processing functions on new mem objects.

        This is useful for parsing other CSV dialects when multiple columns
        convert to a single Chirp column."""

        for attr in dir(mem):
            fname = "_clean_%s" % attr
            if hasattr(self, fname):
                mem = getattr(self, fname)(headers, line, mem)

        return mem

    def _clean_tmode(self, headers, line, mem):
        """ If there is exactly one of [rToneFreq, cToneFreq] columns in the
        csv file, use it for both rtone & ctone. Makes TSQL use friendlier."""

        if self.file_has_rTone and not self.file_has_cTone:
            mem.ctone = mem.rtone
        elif self.file_has_cTone and not self.file_has_rTone:
            mem.rtone = mem.ctone

        return mem

    def _parse_csv_data_line(self, headers, line):
        mem = chirp_common.Memory()
        try:
            if get_datum_by_header(headers, line, "Mode") == "DV":
                mem = chirp_common.DVMemory()
        except OmittedHeaderError:
            pass

        for header in headers:
            try:
                typ, attr = self.ATTR_MAP[header]
            except KeyError:
                continue
            try:
                val = get_datum_by_header(headers, line, header)
                if not val and typ == int:
                    val = None
                else:
                    val = typ(val)
                if hasattr(mem, attr):
                    setattr(mem, attr, val)
            except OmittedHeaderError:
                pass
            except Exception as e:
                raise Exception("[%s] %s" % (attr, e))

        if not mem.power:
            # Default power level to something if not set
            mem.power = DEFAULT_POWER_LEVEL

        return self._clean(headers, line, mem)

    def load(self, filename=None):
        if filename is None and self._filename is None:
            raise errors.RadioError("Need a location to load from")

        if filename:
            self._filename = filename

        self._blank()

        with open(self._filename, newline='', encoding='utf-8-sig') as f:
            return self._load(f)

    def _load(self, f):
        reader = csv.reader(f, delimiter=chirp_common.SEPCHAR, quotechar='"')

        self._comments = []
        good = 0
        lineno = 0
        last_number = -1
        for line in reader:
            # Skip (but stash) comment lines that start with #
            if line and line[0].startswith('#'):
                self._comments.append((last_number, ' '.join(line)))
                continue
            lineno += 1
            if lineno == 1:
                header = line
                self.file_has_rTone = "rToneFreq" in header
                self.file_has_cTone = "cToneFreq" in header
                continue

            if len(header) > len(line):
                LOG.error("Line %i has %i columns, expected %i",
                          lineno, len(line), len(header))
                self.errors.append("Column number mismatch on line %i" %
                                   lineno)
                continue

            try:
                mem = self._parse_csv_data_line(header, line)
                if mem.number is None:
                    raise Exception("Invalid Location field" % lineno)
            except Exception as e:
                LOG.error("Line %i: %s", lineno, e)
                self.errors.append("Line %i: %s" % (lineno, e))
                continue

            last_number = mem.number
            self._grow(mem.number)
            self.memories[mem.number] = mem
            good += 1

        if not good:
            LOG.error(self.errors)
            raise errors.InvalidDataError("No channels found")

    def save(self, filename=None):
        if filename is None and self._filename is None:
            raise errors.RadioError("Need a location to save to")

        if filename:
            self._filename = filename

        with open(self._filename, "w", newline='', encoding='utf-8') as f:
            comments = list(self._comments)
            writer = csv.writer(f, delimiter=chirp_common.SEPCHAR)

            for index, comment in comments[:]:
                if index >= 0:
                    break
                writer.writerow([comment])
                comments.pop(0)

            writer.writerow(chirp_common.Memory.CSV_FORMAT)

            for mem in self.memories:
                for index, comment in comments[:]:
                    if index >= mem.number:
                        break
                    writer.writerow([comment])
                    comments.pop(0)
                write_memory(writer, mem)

    # MMAP compatibility
    def save_mmap(self, filename):
        return self.save(filename)

    def load_mmap(self, filename):
        return self.load(filename)

    def get_memories(self, lo=0, hi=999):
        return [x for x in self.memories if x.number >= lo and x.number <= hi]

    def get_memory(self, number):
        try:
            return self.memories[number].dupe()
        except:
            raise errors.InvalidMemoryLocation("No such memory %s" % number)

    def _grow(self, target):
        delta = target - len(self.memories)
        if delta < 0:
            return

        delta += 1

        for i in range(len(self.memories), len(self.memories) + delta + 1):
            mem = chirp_common.Memory()
            mem.empty = True
            mem.number = i
            self.memories.append(mem)

    def set_memory(self, newmem):
        newmem = newmem.dupe()
        if newmem.power is None:
            newmem.power = DEFAULT_POWER_LEVEL
        else:
            # Accept any power level because we are CSV, but convert it to
            # the class that will str() into our desired format.
            newmem.power = chirp_common.AutoNamedPowerLevel(
                chirp_common.dBm_to_watts(float(newmem.power)))
        self._grow(newmem.number)
        self.memories[newmem.number] = newmem
        self.memories[newmem.number].name = newmem.name.rstrip()

    def erase_memory(self, number):
        mem = chirp_common.Memory()
        mem.number = number
        mem.empty = True
        self.memories[number] = mem

    def get_raw_memory(self, number):
        return ",".join(chirp_common.Memory.CSV_FORMAT) + \
            os.linesep + \
            ",".join(self.memories[number].to_csv())

    @classmethod
    def match_model(cls, filedata, filename):
        """Match files ending in .CSV"""
        try:
            filedata = filedata.decode()
        except UnicodeDecodeError:
            # CSV files are text
            return False
        return filename.lower().endswith("." + cls.FILE_EXTENSION) and \
            (find_csv_header(filedata) or filedata == "")


def find_csv_header(filedata):
    while filedata.startswith('#'):
        filedata = filedata[filedata.find('\n') + 1:]
    return filedata.startswith('Location,')


@directory.register
class CommanderCSVRadio(CSVRadio):
    """A driver for reading CSV files generated by KG-UV Commander software"""
    VENDOR = "Commander"
    MODEL = "KG-UV"
    FILE_EXTENSION = "csv"

    MODE_MAP = {
        "NARR": "NFM",
        "WIDE": "FM",
    }

    SCAN_MAP = {
        "ON":  "",
        "OFF": "S"
    }

    ATTR_MAP = {
        "#":            (int,   "number"),
        "Name":         (str,   "name"),
        "RX Freq":      (chirp_common.parse_freq, "freq"),
        "Scan":         (lambda v: CommanderCSVRadio.SCAN_MAP.get(v), "skip"),
        "TX Dev":       (lambda v: CommanderCSVRadio.MODE_MAP.get(v), "mode"),
        "Group/Notes":  (str,   "comment"),
    }

    def _clean_number(self, headers, line, mem):
        if mem.number == 0:
            for memory in self.memories:
                if memory.empty:
                    mem.number = memory.number
                    break
        return mem

    def _clean_duplex(self, headers, line, mem):
        try:
            txfreq = chirp_common.parse_freq(
                get_datum_by_header(headers, line, "TX Freq"))
        except ValueError:
            mem.duplex = "off"
            return mem

        if mem.freq == txfreq:
            mem.duplex = ""
        elif txfreq:
            mem.duplex = "split"
            mem.offset = txfreq

        return mem

    def _clean_tmode(self, headers, line, mem):
        rtone = get_datum_by_header(headers, line, "Encode")
        ctone = get_datum_by_header(headers, line, "Decode")
        if rtone == "OFF":
            rtone = None
        else:
            rtone = float(rtone)

        if ctone == "OFF":
            ctone = None
        else:
            ctone = float(ctone)

        if rtone:
            mem.tmode = "Tone"
        if ctone:
            mem.tmode = "TSQL"

        mem.rtone = rtone or 88.5
        mem.ctone = ctone or mem.rtone

        return mem

    @classmethod
    def match_model(cls, filedata, filename):
        """Match files ending in .csv and using Commander column names."""
        return filename.lower().endswith("." + cls.FILE_EXTENSION) and \
            filedata.startswith(b"Name,RX Freq,TX Freq,Decode,Encode,TX Pwr,"
                                b"Scan,TX Dev,Busy Lck,Group/Notes") or \
            filedata.startswith(b'"#","Name","RX Freq","TX Freq","Decode",'
                                b'"Encode","TX Pwr","Scan","TX Dev",'
                                b'"Busy Lck","Group/Notes"')


@directory.register
class RTCSVRadio(CSVRadio):
    """A driver for reading CSV files generated by RT Systems software"""
    VENDOR = "RT Systems"
    MODEL = "CSV"
    FILE_EXTENSION = "csv"

    DUPLEX_MAP = {
        "Minus":    "-",
        "Plus":     "+",
        "Simplex":  "",
        "Split":    "split",
    }

    SKIP_MAP = {
        "Off":    "",
        "On":     "S",
        "P Scan": "P",
        "Skip":   "S",
        }

    TMODE_MAP = {
        "None":     "",
        "T Sql":    "TSQL",
    }

    BOOL_MAP = {
        "Off":  False,
        "On":   True,
    }

    ATTR_MAP = {
        "Channel Number":    (int,   "number"),
        "Receive Frequency": (chirp_common.parse_freq, "freq"),
        "Offset Frequency":  (chirp_common.parse_freq, "offset"),
        "Offset Direction":  (lambda v:
                              RTCSVRadio.DUPLEX_MAP.get(v, v), "duplex"),
        "Operating Mode":    (str,   "mode"),
        "Name":              (str,   "name"),
        "Tone Mode":         (lambda v:
                              RTCSVRadio.TMODE_MAP.get(v, v), "tmode"),
        "CTCSS":             (lambda v:
                              float(v.split(" ")[0]), "rtone"),
        "DCS":               (int,   "dtcs"),
        "Skip":              (lambda v:
                              RTCSVRadio.SKIP_MAP.get(v, v), "skip"),
        "Step":              (lambda v:
                              float(v.split(" ")[0]), "tuning_step"),
        "Mask":              (lambda v:
                              RTCSVRadio.BOOL_MAP.get(v, v), "empty",),
        "Comment":           (str,   "comment"),
        }

    def _clean_duplex(self, headers, line, mem):
        if mem.duplex == "split":
            try:
                val = get_datum_by_header(headers, line, "Transmit Frequency")
                val = chirp_common.parse_freq(val)
                mem.offset = val
            except OmittedHeaderError:
                pass

        return mem

    def _clean_mode(self, headers, line, mem):
        if mem.mode == "FM":
            try:
                val = get_datum_by_header(headers, line, "Half Dev")
                if self.BOOL_MAP[val]:
                    mem.mode = "FMN"
            except OmittedHeaderError:
                pass

        return mem

    def _clean_ctone(self, headers, line, mem):
        # RT Systems only stores a single tone value
        mem.ctone = mem.rtone
        return mem

    @classmethod
    def match_model(cls, filedata, filename):
        """Match files ending in .csv and using RT Systems column names."""
        # RT Systems provides a different set of columns for each radio.
        # We attempt to match only the first few columns, hoping they are
        # consistent across radio models.
        try:
            filedata = filedata.decode()
        except UnicodeDecodeError:
            # CSV files are text
            return False
        return filename.lower().endswith("." + cls.FILE_EXTENSION) and \
            filedata.startswith("Channel Number,Receive Frequency,"
                                "Transmit Frequency,Offset Frequency,"
                                "Offset Direction,Operating Mode,"
                                "Name,Tone Mode,CTCSS,DCS")



# Copyright 2022 Dan Smith <chirp@f.danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import atexit
import contextlib
import functools
import itertools
import logging
import os
import platform
import shutil
import tempfile
import threading

import wx

from chirp import chirp_common
from chirp.drivers import generic_csv
from chirp import errors
from chirp import logger
from chirp import platform as chirp_platform
from chirp import settings
from chirp.wxui import config
from chirp.wxui import radiothread

LOG = logging.getLogger(__name__)
CONF = config.get()

CHIRP_DATA_MEMORY = wx.DataFormat('x-chirp/memory-channel')
EditorChanged, EVT_EDITOR_CHANGED = wx.lib.newevent.NewCommandEvent()
StatusMessage, EVT_STATUS_MESSAGE = wx.lib.newevent.NewCommandEvent()
INDEX_CHAR = settings.BANNED_NAME_CHARACTERS[0]

# This is a lock that can be used to exclude edit-specific operations
# from happening at the same time, like radiothread async
# operations. This needs to be local to the thing with the wx.Grid as
# its child, which is the chirp main window in our case. Making this
# global is technically too broad, but in reality, it's equivalent for
# us at the moment, and this is easier.
EDIT_LOCK = threading.Lock()


def closes_clipboard(fn):
    @functools.wraps(fn)
    def wrapper(*a, **k):
        try:
            return fn(*a, **k)
        finally:
            if wx.TheClipboard.IsOpened():
                LOG.warning('Closing clipboard left open by %s' % fn)
                wx.TheClipboard.Close()
    return wrapper


class LiveAdapter(generic_csv.CSVRadio):
    FILE_EXTENSION = 'img'

    def __init__(self, liveradio):
        # Python2 old-style class compatibility
        generic_csv.CSVRadio.__init__(self, None)
        self._liveradio = liveradio
        self.VENDOR = liveradio.VENDOR
        self.MODEL = liveradio.MODEL
        self.VARIANT = liveradio.VARIANT
        self.BAUD_RATE = liveradio.BAUD_RATE
        self.HARDWARE_FLOW = liveradio.HARDWARE_FLOW
        self.pipe = liveradio.pipe
        self._features = self._liveradio.get_features()

    def get_features(self):
        return self._features

    def set_pipe(self, pipe):
        self.pipe = pipe
        self._liveradio.pipe = pipe

    def sync_in(self):
        for i in range(*self._features.memory_bounds):
            mem = self._liveradio.get_memory(i)
            self.set_memory(mem)
            status = chirp_common.Status()
            status.max = self._features.memory_bounds[1]
            status.cur = i
            status.msg = 'Cloning'
            self.status_fn(status)

    def sync_out(self):
        # FIXME: Handle errors
        for i in range(*self._features.memory_bounds):
            mem = self.get_memory(i)
            if mem.freq == 0:
                # Convert the CSV notion of emptiness
                try:
                    self._liveradio.erase_memory(i)
                except errors.RadioError as e:
                    LOG.error(e)
            else:
                try:
                    self._liveradio.set_memory(mem)
                except errors.RadioError as e:
                    LOG.error(e)
            status = chirp_common.Status()
            status.max = self._features.memory_bounds[1]
            status.cur = i
            status.msg = _('Cloning')
            self.status_fn(status)

    def get_settings(self):
        return self._liveradio.get_settings()

    def set_settings(self, settings):
        return self._liveradio.set_settings(settings)


class EditorMenuItem(wx.MenuItem):
    MENU_VIEW = 'View'
    MENU_EDIT = 'Edit'
    ITEMS = {}

    def __init__(self, cls, callback_name, *a, **k):
        self._wx_id = wx.NewId()
        super().__init__(None, self._wx_id, *a, **k)
        self._cls = cls
        self._callback_name = callback_name
        self.ITEMS[self.key] = self._wx_id

    @property
    def key(self):
        return '%s:%s' % (self._cls.__name__, self._callback_name)

    @property
    def editor_class(self):
        return self._cls

    def editor_callback(self, editor, event):
        getattr(editor, self._callback_name)(event)

    def add_menu_callback(self):
        # Some platforms won't actually set the accelerator (macOS) or allow
        # enable/check (linux) before we are in a menu.
        accel = self.GetAccel()
        self.SetAccel(None)
        self.SetAccel(accel)


class EditorMenuItemToggle(EditorMenuItem):
    """An EditorMenuItem that manages boolean/check state in CONF"""
    def __init__(self, cls, callback_name, conf_tuple, *a, **k):
        k['kind'] = wx.ITEM_CHECK
        super().__init__(cls, callback_name, *a, **k)
        self._conf_key, self._conf_section = conf_tuple

    def editor_callback(self, editor, event):
        menuitem = event.GetEventObject().FindItemById(event.GetId())
        CONF.set_bool(self._conf_key, menuitem.IsChecked(), self._conf_section)
        super().editor_callback(editor, event)

    def add_menu_callback(self):
        super().add_menu_callback()
        self.Check(CONF.get_bool(self._conf_key, self._conf_section, False))


class ChirpEditor(wx.Panel):
    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self.setup_radio_interface()
        self.wait_dialog = None

    @property
    def radio(self):
        return self._radio

    def start_wait_dialog(self, message):
        if self.wait_dialog:
            LOG.error('Wait dialog already in progress!')
            return

        self.wait_dialog = wx.ProgressDialog(_('Please wait'), message, 100,
                                             parent=self)
        wx.CallAfter(self.wait_dialog.Show)

    def bump_wait_dialog(self, value=None, message=None):
        if value:
            wx.CallAfter(self.wait_dialog.Update, value, newmsg=message)
        else:
            wx.CallAfter(self.wait_dialog.Pulse, message)

    def stop_wait_dialog(self):
        def cb():
            if self.wait_dialog:
                self.wait_dialog.Destroy()
                self.wait_dialog = None

        wx.CallAfter(cb)

    def status_message(self, message):
        wx.PostEvent(self, StatusMessage(self.GetId(), message=message))

    def refresh(self):
        pass

    def cb_copy(self, cut=False):
        raise NotImplementedError()

    @closes_clipboard
    def cb_copy_data(self, data):
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(data)
            wx.TheClipboard.Close()
        else:
            raise RuntimeError(_('Unable to open the clipboard'))

    @closes_clipboard
    def cb_paste(self):
        memdata = wx.CustomDataObject(CHIRP_DATA_MEMORY)
        textdata = wx.TextDataObject()
        if wx.TheClipboard.Open():
            gotchirpmem = wx.TheClipboard.GetData(memdata)
            got = wx.TheClipboard.GetData(textdata)
            wx.TheClipboard.Close()
        if gotchirpmem:
            return memdata
        elif got:
            return textdata

    def cb_delete(self):
        pass

    def cb_find(self, text):
        pass

    def select_all(self):
        pass

    def saved(self):
        pass

    def selected(self):
        pass

    def get_scroll_pos(self):
        return None

    def set_scroll_pos(self, pos):
        LOG.warning('Editor %s does not support set_scroll_pos()' % (
            self.__class__.__name__))

    def update_font(self):
        pass

    @classmethod
    def get_menu_items(self):
        """Return a dict of menu items specific to this editor class

        Example: {'Edit': [wx.MenuItem], 'View': [wx.MenuItem]}
        """
        return {}


class ChirpSyncEditor:
    """Radio interface that makes synchronous calls.

    This executes RadioJob requests synchronously in the calling
    thread and directly spawns the callback. To be used only with
    radio drivers that manipulate in-memory state.
    """
    def do_radio(self, cb, fn, *a, **k):
        """Synchronous passthrough for non-Live radios"""
        job = radiothread.RadioJob(self, fn, a, k)
        try:
            job.result = getattr(self._radio, fn)(*a, **k)
        except Exception as e:
            LOG.exception('Failed to run %s(%s, %s)' % (
                fn, ','.join(str(x) for x in a),
                ','.join('%s=%r' % (k, v) for k, v in k.items())))
            job.result = e
        if cb:
            cb(job)

    do_lazy_radio = do_radio

    def setup_radio_interface(self):
        pass

    @property
    def busy(self):
        return False


class ChirpAsyncEditor(ChirpSyncEditor):
    """Radio interface that makes async calls in a helper thread.

    This executes RadioJob requests asynchronously in a thread and
    schedules the callback via a wx event.

    """
    def do_radio(self, cb, fn, *a, **k):
        self._jobs[self._radio_thread.submit(self, fn, *a, **k)] = cb

    def do_lazy_radio(self, cb, fn, *a, **k):
        self._jobs[self._radio_thread.background(self, fn, *a, **k)] = cb

    def radio_thread_event(self, job, block=True):
        if job.fn == 'get_memory':
            msg = _('Refreshed memory %s') % job.args[0]
        elif job.fn == 'set_memory':
            msg = _('Uploaded memory %s') % job.args[0].number
        elif job.fn == 'get_settings':
            msg = _('Retrieved settings')
        elif job.fn == 'set_settings':
            msg = _('Saved settings')
        elif job.fn == 'erase_memory':
            msg = _('Erased memory %s') % job.args[0]
        else:
            msg = _('Finished radio job %s') % job.fn

        if not EDIT_LOCK.acquire(block):
            return False
        try:
            self.status_message(msg)
            cb = self._jobs.pop(job.id)
            if cb:
                wx.CallAfter(cb, job)

            # Update our status, which may be reset to unmodified
            wx.PostEvent(self, EditorChanged(self.GetId()))
        finally:
            EDIT_LOCK.release()

        return True

    def set_radio_thread(self, radio_thread):
        self._radio_thread = radio_thread

    def setup_radio_interface(self):
        self._jobs = {}

    @property
    def busy(self):
        return self._radio_thread.pending != 0


class ChirpSettingGrid(wx.Panel):
    def __init__(self, settinggroup, *a, **k):
        super(ChirpSettingGrid, self).__init__(*a, **k)
        self._group = settinggroup
        self._settings = {}
        self._needs_reload = False

        self.pg = wx.propgrid.PropertyGrid(
            self,
            style=wx.propgrid.PG_SPLITTER_AUTO_CENTER |
            wx.propgrid.PG_BOLD_MODIFIED)

        self.pg.Bind(wx.propgrid.EVT_PG_CHANGED, self._pg_changed)
        self.pg.Bind(wx.EVT_MOTION, self._mouseover)

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)
        sizer.Add(self.pg, 1, wx.EXPAND)

        self._choices = {}

        self._add_items(self._group)
        self.pg.Bind(wx.propgrid.EVT_PG_CHANGING, self._check_change)

    def _mouseover(self, event):
        prop = self.pg.HitTest(event.GetPosition()).GetProperty()
        tip = None
        if prop:
            setting = self.get_setting_by_name(prop.GetName())
            tip = setting.__doc__ or None
            if tip and tip == setting.get_name():
                # This is the default, which makes no sense, but it's been
                # that way for ages, so just avoid exposing it here if it's
                # set to the default.
                tip = None

        event.GetEventObject().SetToolTip(tip)

    def _add_items(self, group, parent=None):
        def append(item):
            if parent:
                self.pg.AppendIn(parent, item)
            else:
                self.pg.Append(item)

        for name, element in group.items():
            if isinstance(element, settings.RadioSettingSubGroup):
                category = wx.propgrid.PropertyCategory(
                               element.get_shortname(), element.get_name())
                append(category)
                self._add_items(element, parent=category)
                continue
            elif not isinstance(element, settings.RadioSetting):
                LOG.debug('Skipping nested group %s' % element)
                continue
            if len(element.keys()) > 1:
                append(wx.propgrid.PropertyCategory(
                    element.get_shortname()))

            for i in element.keys():
                value = element[i]
                if isinstance(value, settings.RadioSettingValueInteger):
                    editor = self._get_editor_int(element, value)
                elif isinstance(value, settings.RadioSettingValueFloat):
                    editor = self._get_editor_float(element, value)
                elif isinstance(value, settings.RadioSettingValueList):
                    editor = self._get_editor_choice(element, value)
                elif isinstance(value, settings.RadioSettingValueBoolean):
                    editor = self._get_editor_bool(element, value)
                elif isinstance(value, settings.RadioSettingValueString):
                    editor = self._get_editor_str(element, value)
                else:
                    LOG.warning('Unsupported setting type %r' % value)
                    editor = None
                if not editor:
                    continue

                editor.SetName('%s%s%i' % (name, INDEX_CHAR, i))
                if len(element.keys()) > 1:
                    editor.SetLabel('')
                editor.Enable(value.get_mutable())
                append(editor)
                self._settings[element.get_name()] = element
                if editor.IsValueUnspecified():
                    # Mark invalid/unspecified values so the user can fix them
                    self.pg.SetPropertyBackgroundColour(editor.GetName(),
                                                        wx.YELLOW)

        self.pg.Bind(wx.propgrid.EVT_PG_CHANGING, self._check_change)

    def get_setting_by_name(self, name, index=0):
        if INDEX_CHAR in name:
            # FIXME: This will only work for single-index settings of course
            name, _ = name.split(INDEX_CHAR, 1)
        for setting_name, setting in self._settings.items():
            if name == setting_name:
                return setting

    def _check_change(self, event):
        setting = self.get_setting_by_name(event.GetPropertyName())
        if not setting:
            LOG.error('Got change event for unknown setting %s' % (
                event.GetPropertyName()))
            return
        warning = setting.get_warning(event.GetValue())
        if warning:
            r = wx.MessageBox(warning, _('WARNING!'),
                              wx.OK | wx.CANCEL | wx.CANCEL_DEFAULT)
            if r == wx.CANCEL:
                LOG.info('User aborted setting %s=%s with warning message',
                         event.GetPropertyName(), event.GetValue())
                event.SetValidationFailureBehavior(0)
                event.Veto()
                return
            else:
                LOG.info('User made change to %s=%s despite warning',
                         event.GetPropertyName(), event.GetValue())
        self._needs_reload = setting.volatile
        if self.needs_reload:
            wx.MessageBox(_(
                'Changing this setting requires refreshing the settings from '
                'the image, which will happen now.'),
                          _('Refresh required'), wx.OK)

        # If we were unspecified or otherwise marked, clear those markings
        self.pg.SetPropertyColoursToDefault(event.GetProperty().GetName())

    @property
    def name(self):
        return self._group.get_name()

    @property
    def needs_reload(self):
        return self._needs_reload

    @property
    def propgrid(self):
        return self.pg

    def _pg_changed(self, event):
        wx.PostEvent(self, EditorChanged(self.GetId()))

    def _get_editor_int(self, setting, value):
        e = wx.propgrid.IntProperty(setting.get_shortname(),
                                    setting.get_name())
        if value.initialized:
            e.SetValue(int(value))
        else:
            e.SetValueToUnspecified()
        e.SetEditor('SpinCtrl')
        e.SetAttribute(wx.propgrid.PG_ATTR_MIN, value.get_min())
        e.SetAttribute(wx.propgrid.PG_ATTR_MAX, value.get_max())
        e.SetAttribute(wx.propgrid.PG_ATTR_SPINCTRL_STEP, value.get_step())
        return e

    def _get_editor_float(self, setting, value):
        class ChirpFloatProperty(wx.propgrid.FloatProperty):
            def ValidateValue(self, val, info):
                if value.get_min() and not val >= value.get_min():
                    info.SetFailureMessage(
                        _('Value must be at least %.4f' % value.get_min()))
                    return False
                if value.get_max() and not val <= value.get_max():
                    info.SetFailureMessage(
                        _('Value must be at most %.4f' % value.get_max()))
                    return False
                return True

            def ValueToString(self, _value, flags=0):
                return value.format(_value)

        e = ChirpFloatProperty(setting.get_shortname(),
                               setting.get_name())
        if value.initialized:
            e.SetValue(float(value))
        else:
            e.SetValueToUnspecified()
        return e

    def _get_editor_choice(self, setting, value):
        choices = value.get_options()
        self._choices[setting.get_name()] = choices
        current = choices.index(str(value))
        return wx.propgrid.EnumProperty(setting.get_shortname(),
                                        setting.get_name(),
                                        choices, range(len(choices)),
                                        current)

    def _get_editor_bool(self, setting, value):
        prop = wx.propgrid.BoolProperty(setting.get_shortname(),
                                        setting.get_name(),
                                        bool(value))
        prop.SetAttribute(wx.propgrid.PG_BOOL_USE_CHECKBOX, True)
        return prop

    def _get_editor_str(self, setting, value):
        class ChirpStrProperty(wx.propgrid.StringProperty):
            def ValidateValue(self, text, info):
                try:
                    value.set_value(text)
                except Exception as e:
                    info.SetFailureMessage(str(e))
                    return False
                return True

        e = ChirpStrProperty(setting.get_shortname(),
                             setting.get_name())
        if value.initialized:
            e.SetValue(str(value))
        else:
            e.SetValueToUnspecified()
        return e

    def get_setting_values(self):
        """Return a dict of {name: (RadioSetting, newvalue)}"""
        values = {}
        for prop in self.pg._Items():
            if prop.IsCategory():
                continue
            if prop.IsValueUnspecified():
                continue
            basename = prop.GetName().split(INDEX_CHAR)[0]
            if isinstance(prop, wx.propgrid.EnumProperty):
                value = self._choices[basename][prop.GetValue()]
            else:
                value = prop.GetValue()
            setting = self._settings[basename]
            values[prop.GetName()] = setting, value
        return values

    def get_values(self):
        """Return a dict of {name: newvalue}"""
        return {k: v[1] for k, v in self.get_setting_values().items()}

    def saved(self):
        for prop in self.pg._Items():
            prop.SetModifiedStatus(False)


def _error_proof(*expected_errors):
    """Decorate a method and display an error if it raises.

    If the method raises something in expected_errors, then
    log an error, otherwise log exception.
    """

    def show_error(msg):
        d = wx.MessageDialog(None, str(msg), _('An error has occurred'),
                             style=wx.OK | wx.ICON_ERROR)
        d.ShowModal()

    def wrap(fn):
        @functools.wraps(fn)
        def inner(*args, **kwargs):
            try:
                return fn(*args, **kwargs)
            except expected_errors as e:
                LOG.error('%s: %s' % (fn, e))
                show_error(e)
            except Exception as e:
                LOG.exception('%s raised unexpected exception' % fn)
                show_error(e)

        return inner
    return wrap


class error_proof(object):
    def __init__(self, *expected_exceptions):
        self._expected = expected_exceptions
        self.fn = None

    @staticmethod
    def show_error(msg):
        d = wx.MessageDialog(None, str(msg), _('An error has occurred'),
                             style=wx.OK | wx.ICON_ERROR)
        d.ShowModal()

    def run_safe(self, fn, args, kwargs):
        try:
            return fn(*args, **kwargs)
        except self._expected as e:
            LOG.error('%s: %s' % (fn, e))
            self.show_error(e)
        except Exception as e:
            LOG.exception('%s raised unexpected exception' % fn)
            self.show_error(e)

    def __call__(self, fn):
        self.fn = fn

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            return self.run_safe(fn, args, kwargs)
        return wrapper

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, traceback):
        if exc_type:
            if exc_type in self._expected:
                LOG.error('%s: %s: %s',
                          self.fn or 'context', exc_type, exc_val)
                self.show_error(exc_val)
                return True
            else:
                LOG.exception('Context raised unexpected_exception',
                              exc_info=(exc_type, exc_val, traceback))
                self.show_error(exc_val)


def reveal_location(path):
    LOG.debug('Revealing path %s', path)
    system = platform.system()
    if system == 'Windows':
        if not os.path.isdir(path):
            # Windows can only reveal the containing directory of a file
            path = os.path.dirname(path)
        wx.Execute('explorer %s' % path)
    elif system == 'Darwin':
        wx.Execute('open -R %s' % path)
    elif system == 'Linux':
        wx.Execute('open %s' % path)
    else:
        raise Exception(_('Unable to reveal %s on this system') % path)


def delete_atexit(path_):
    def do(path):
        try:
            os.remove(path)
            LOG.debug('Removed temporary file %s', path)
        except Exception as e:
            LOG.warning('Failed to remove %s: %s', path, e)

    atexit.register(do, path_)


def temporary_debug_log():
    """Return a temporary copy of our debug log"""
    pf = chirp_platform.get_platform()
    src = pf.config_file('debug.log')
    fd, dst = tempfile.mkstemp(
        prefix='chirp_debug-',
        suffix='.txt')
    delete_atexit(dst)
    shutil.copy(src, dst)
    return dst


@contextlib.contextmanager
def expose_logs(level, root, label):
    if not isinstance(root, tuple):
        root = (root,)

    mgrs = (logger.log_history(level, x) for x in root)
    with contextlib.ExitStack() as stack:
        histories = [stack.enter_context(m) for m in mgrs]
        try:
            yield
        finally:
            lines = list(itertools.chain.from_iterable(x.get_history()
                                                       for x in histories))
            if lines:
                msg = os.linesep.join(x.getMessage() for x in lines)
                d = wx.MessageDialog(
                    None, str(msg), label,
                    style=wx.OK | wx.ICON_INFORMATION)
                d.ShowModal()



# Copyright 2015 David Fannin KK6DF  <kk6df@arrl.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
import logging

from chirp import bitwise
from chirp import chirp_common
from chirp import directory
from chirp import errors
from chirp import memmap
from chirp import util
from chirp.settings import RadioSettingGroup, RadioSetting, RadioSettings, \
    RadioSettingValueList, RadioSettingValueString, RadioSettingValueInteger

LOG = logging.getLogger(__name__)

#
#  Chirp Driver for TYT TH-9000D (models: 2M (144 MHz), 1.25M (220 MHz)
#                                 and 70 cm (440 MHz) radios)
#
#  Version 1.0
#
#         - Skip channels
#
# Global Parameters
#
MMAPSIZE = 16384
TONES = [62.5] + list(chirp_common.TONES)
TMODES = ['', 'Tone', 'DTCS', '']
DUPLEXES = ['', 'err', '-', '+']  # index 2 not used
MODES = ['WFM', 'FM', 'NFM']  # 25k, 20k,15k bw
TUNING_STEPS = [5.0, 6.25, 8.33, 10.0, 12.5, 15.0, 20.0, 25.0, 30.0, 50.0]
#               index 0-9
POWER_LEVELS = [chirp_common.PowerLevel("High", watts=65),
                chirp_common.PowerLevel("Mid", watts=25),
                chirp_common.PowerLevel("Low", watts=10)]

CROSS_MODES = chirp_common.CROSS_MODES

APO_LIST = ["Off", "30 min", "1 hr", "2 hrs"]
BGCOLOR_LIST = ["Blue", "Orange", "Purple"]
BGBRIGHT_LIST = ["%s" % x for x in range(1, 32)]
SQUELCH_LIST = ["Off"] + ["Level %s" % x for x in range(1, 20)]
TIMEOUT_LIST = ["Off"] + ["%s min" % x for x in range(1, 30)]
TXPWR_LIST = ["60W", "25W"]  # maximum power for Hi setting
TBSTFREQ_LIST = ["1750 Hz", "2100 Hz", "1000 Hz", "1450 Hz"]
BEEP_LIST = ["Off", "On"]

MEM_FORMAT = """
#seekto 0x0000;
struct {
   u8 unknown0000[16];
   char idhdr[16];
   u8 unknown0001[16];
} fidhdr;
"""
# Overall Memory Map:
#
#    Memory Map (Range 0x0100-3FF0, step 0x10):
#
#        Field                   Start  End  Size
#                                (hex)  (hex) (hex)
#
#        1 Channel Set Flag        0100  011F   20
#        2 Channel Skip Flag       0120  013F   20
#        3 Blank/Unknown           0140  01EF   B0
#        4 Unknown                 01F0  01FF   10
#        5 TX/RX Range             0200  020F   10
#        6 Bootup Passwd           0210  021F   10
#        7 Options, Radio          0220  023F   20
#        8 Unknown                 0240  019F
#            8B Startup Label      03E0  03E7   07
#        9 Channel Bank            2000  38FF 1900
#             Channel 000          2000  201F   20
#             Channel 001          2020  202F   20
#             ...
#             Channel 199          38E0  38FF   20
#        10 Blank/Unknown          3900  3FFF  6FF  14592  16383    1792
#            Total Map Size           16128 (2^8 = 16384)
#
#  TH9000/220  memory map
#  section: 1 and 2:  Channel Set/Skip Flags
#
#    Channel Set (starts 0x100) : Channel Set  bit is value 0 if a memory
#                                 location in the channel bank is active.
#    Channel Skip (starts 0x120): Channel Skip bit is value 0 if a memory
#                                 location in the channel bank is active.
#
#    Both flag maps are a total 24 bytes in length, aligned on 32 byte records.
#    bit = 0 channel set/no skip,  1 is channel not set/skip
#
#    to index a channel:
#        cbyte = channel / 8 ;
#        cbit  = channel % 8 ;
#        setflag  = csetflag[cbyte].c[cbit] ;
#        skipflag = cskipflag[cbyte].c[cbit] ;
#
#    channel range is 0-199, range is 32 bytes (last 7 unknown)
#
MEM_FORMAT = MEM_FORMAT + """
#seekto 0x0100;
struct {
   bit c[8];
} csetflag[32];

struct {
   u8 unknown0100[7];
} ropt0100;

#seekto 0x0120;
struct {
   bit c[8];
} cskipflag[32];

struct {
   u8 unknown0120[7];
} ropt0120;
"""
#  TH9000  memory map
#  section: 5  TX/RX Range
#     used to set the TX/RX range of the radio (e.g. 222-228 MHz for 220 meter)
#     possible to set range for tx/rx
#
MEM_FORMAT = MEM_FORMAT + """
#seekto 0x0200;
struct {
    bbcd txrangelow[4];
    bbcd txrangehi[4];
    bbcd rxrangelow[4];
    bbcd rxrangehi[4];
} freqrange;
"""
# TH9000  memory map
# section: 6  bootup_passwd
#    used to set bootup passwd (see boot_passwd checkbox option)
#
#  options - bootup password
#
#  bytes:bit   type                 description
#  ---------------------------------------------------------------------------
#  6         u8 bootup_passwd[6]     bootup passwd, 6 chars, numeric chars
#                                    30-39 , see boot_passwd checkbox to set
#  10        u8 unknown;
#

MEM_FORMAT = MEM_FORMAT + """
#seekto 0x0210;
struct {
   u8 bootup_passwd[6];
   u8 unknown2010[10];
} ropt0210;
"""
#  TH9000/220  memory map
#  section: 7  Radio Options
#        used to set a number of radio options
#
#  bytes:bit   type                 description
#  ---------------------------------------------------------------------------
#  1         u8 display_mode     display mode, range 0-2, 0=freq,1=channel,
#                                2=name (selecting name affects vfo_mr)
#  1         u8 vfo_mr;          vfo_mr , 0=vfo, mr=1
#  1         u8 unknown;
#  1         u8 squelch;         squelch level, range 0-19, hex for menu
#  1         u8 unknown[2];
#  1         u8 channel_lock;    if display_mode[channel] selected, then
#                                lock=1,no lock =0
#  1         u8 unknown;
#  1         u8 bg_brightness ;  background brightness, range 0-21, hex,
#                                menu index
#  1         u8 unknown;
#  1         u8 bg_color ;       bg color, menu index,  blue 0 , orange 1,
#                                purple 2
#  1         u8 tbst_freq ;      tbst freq, menu 0=1750 Hz, 1=2100,
#                                2=1000, 3=1450 Hz
#  1         u8 timeout_timer;   timeout timer, hex, value = minutes,
#                                0= no timeout
#  1         u8 unknown;
#  1         u8 auto_power_off;   auto power off, range 0-3, off,30min, 1hr,
#                                 2hr, hex menu index
#  1         u8 voice_prompt;     voice prompt, value 0,1 , Beep ON = 1,
#                                 Beep Off = 2
#
# description of function setup options, starting at 0x0230
#
#  bytes:bit   type                 description
#  ---------------------------------------------------------------------------
#  1         u8  // 0
#   :4       unknown:6
#   :1       elim_sql_tail:1   eliminate squelsh tail when no ctcss checkbox
#                              (1=checked)
#   :1       sql_key_function  "squelch off" 1 , "squelch momentary off" 0 ,
#                              menu index
#  2         u8 unknown[2] /1-2
#  1         u8 // 3
#   :4       unknown:4
#   :1       inhibit_init_ops:1 //bit 5
#   :1       unknownD:1
#   :1       inhibit_setup_bg_chk:1 //bit 7
#   :1       unknown:1
#  1         u8 tail_elim_type    menu , (off=0,120=1,180=2),  // 4
#  1         u8 choose_tx_power    menu , (60w=0,25w=1) // 5
#  2         u8 unknown[2]; // 6-7
#  1         u8 bootup_passwd_flag  checkbox 1=on, 0=off // 8
#  7         u8 unknown[7]; // 9-F
#
MEM_FORMAT = MEM_FORMAT + """
#seekto 0x0220;
struct {
   u8 display_mode;
   u8 vfo_mr;
   u8 unknown0220A;
   u8 squelch;
   u8 unknown0220B[2];
   u8 channel_lock;
   u8 unknown0220C;
   u8 bg_brightness;
   u8 unknown0220D;
   u8 bg_color;
   u8 tbst_freq;
   u8 timeout_timer;
   u8 unknown0220E;
   u8 auto_power_off;
   u8 voice_prompt;
   u8 unknown0230A:6,
      elim_sql_tail:1,
      sql_key_function:1;
   u8 unknown0230B[2];
   u8 unknown0230C:4,
      inhibit_init_ops:1,
      unknown0230D:1,
      inhibit_setup_bg_chk:1,
      unknown0230E:1;
   u8 tail_elim_type;
   u8 choose_tx_power;
   u8 unknown0230F[2];
   u8 bootup_passwd_flag;
   u8 unknown0230G[7];
} settings;
"""
#  TH9000  memory map
#  section: 8B  Startup Label
#
#  bytes:bit   type                 description
#  ---------------------------------------------------------------------------
#  7     char start_label[7]    label displayed at startup (usually your call
#                               sign)
#
MEM_FORMAT = MEM_FORMAT + """
#seekto 0x03E0;
struct {
    char startname[7];
} slabel;
"""
#  TH9000/220  memory map
#  section: 9  Channel Bank
#         description of channel bank (200 channels , range 0-199)
#         Each 32 Byte (0x20 hex)  record:
#  bytes:bit   type                 description
#  ---------------------------------------------------------------------------
#  4         bbcd freq[4]        receive frequency in packed binary coded
#                                decimal
#  4         bbcd offset[4]      transmit offset in packed binary coded decimal
#                                (note: plus/minus direction set by 'duplex'
#                                field)
#  1         u8
#   :4       unknown:4
#   :4       tuning_step:4         tuning step, menu index value from 0-9
#            5,6.25,8.33,10,12.5,15,20,25,30,50
#  1         u8
#   :4       unknown:4          not yet decoded, used for DCS coding?
#   :2       channel_width:2     channel spacing, menu index value from 0-3
#            25,20,12.5
#   :1       reverse:1           reverse flag, 0=off, 1=on (reverses tx and
#                                rx freqs)
#   :1       txoff:1             transmitt off flag, 0=transmit ,
#                                1=do not transmit
#  1         u8
#   :1       talkaround:1        talkaround flag, 0=off, 1=on (bypasses
#                                repeater)
#   :1       compander:1         compander flag, 0=off, 1=on (turns on/off
#                                voice compander option)
#   :2       unknown:2
#   :2       power:2             tx power setting, value range 0-2, 0=hi,
#                                1=med, 2=lo
#   :2       duplex:2            duplex settings, 0=simplex,
#                                2= minus(-) offset, 3= plus (+) offset
#                               (see offset field)
#
#  1         u8
#   :4       unknown:4
#   :2       rxtmode:2           rx tone mode, value range 0-2, 0=none,
#                                1=CTCSS, 2=DCS  (ctcss tone in field rxtone)
#   :2       txtmode:2           tx tone mode, value range 0-2, 0=none,
#                                1=CTCSS, 3=DCS  (ctcss tone in field txtone)
#  1         u8
#   :2       unknown:2
#   :6       txtone:6            tx ctcss tone, menu index
#  1         u8
#   :2       unknown:2
#   :6       rxtone:6            rx ctcss tone, menu index
#  1         u8 txcode           ?, not used for ctcss
#  1         u8 rxcode           ?, not used for ctcss
#  3         u8 unknown[3]
#  7         char name[7]        7 byte char string for channel name
#  1         u8
#   :6       unknown:6,
#   :2       busychannellockout:2 busy channel lockout option ,
#                                 0=off, 1=repeater, 2=busy
#                                 (lock out tx if channel busy)
#  4         u8 unknownI[4];
#  1         u8
#   :7       unknown:7
#   :1       scrambler:1         scrambler flag, 0=off, 1=on (turns on tyt
#                                scrambler option)
#
MEM_FORMAT = MEM_FORMAT + """
#seekto 0x2000;
struct {
  bbcd freq[4];
  bbcd offset[4];
  u8 unknown2000A:4,
     tuning_step:4;
  u8 rxdcsextra:1,
     txdcsextra:1,
     rxinv:1,
     txinv:1,
     channel_width:2,
     reverse:1,
     txoff:1;
  u8 talkaround:1,
     compander:1,
     unknown2000C:2,
     power:2,
     duplex:2;
  u8 unknown2000D:4,
     rxtmode:2,
     txtmode:2;
  u8 unknown2000E:2,
     txtone:6;
  u8 unknown2000F:2,
     rxtone:6;
  u8 txcode;
  u8 rxcode;
  u8 unknown2000G[3];
  char name[7];
  u8 unknown2000H:6,
     busychannellockout:2;
  u8 unknown2000I[4];
  u8 unknown2000J:7,
     scrambler:1;
} memory[200] ;
"""


def _echo_write(radio, data):
    try:
        radio.pipe.write(data)
        radio.pipe.read(len(data))
    except Exception as e:
        LOG.error("Error writing to radio: %s" % e)
        raise errors.RadioError("Unable to write to radio")


def _checksum(data):
    cs = 0
    for byte in data:
        cs += byte
    return cs % 256


def _read(radio, length):
    try:
        data = radio.pipe.read(length)
    except Exception as e:
        LOG.error("Error reading from radio: %s" % e)
        raise errors.RadioError("Unable to read from radio")

    if len(data) != length:
        LOG.error("Short read from radio (%i, expected %i)" % (len(data),
                  length))
        LOG.debug(util.hexprint(data))
        raise errors.RadioError("Short read from radio")
    return data


def _ident(radio):
    radio.pipe.timeout = 1
    exito = False
    for i in range(0, 5):
        _echo_write(radio, b"PROGRAM")
        response = radio.pipe.read(3)

        if response == b"QX\06":
            exito = True
            break

    # check if we had EXITO
    if exito is False:
        msg = "The radio did not accept program mode after five tries.\n"
        msg += "Check you interface cable and power cycle your radio."
        raise errors.RadioError(msg)

    _echo_write(radio, b"\x02")
    response = radio.pipe.read(16)
    LOG.debug(util.hexprint(response))
    if response[1:8] != b"TH-9000":
        LOG.error("Looking  for:\n%s" % util.hexprint("TH-9000"))
        LOG.error("Response was:\n%s" % util.hexprint(response))
        raise errors.RadioError("Unsupported model")


def _send(radio, cmd, addr, length, data=None):
    frame = struct.pack(">cHb", cmd, addr, length)
    if data:
        frame += data
        frame += bytes([_checksum(frame[1:])])
        frame += b"\x06"
    _echo_write(radio, frame)
    LOG.debug("Sent:\n%s" % util.hexprint(frame))
    if data:
        result = radio.pipe.read(1)
        if result != b"\x06":
            LOG.debug("Ack was: %s" % repr(result))
            raise errors.RadioError(
                "Radio did not accept block at %04x" % addr)
        return
    result = _read(radio, length + 6)
    LOG.debug("Got:\n%s" % util.hexprint(result))
    header = result[:4]
    data = result[4:-2]
    ack = result[-1:]
    if ack != b"\x06":
        LOG.debug("Ack was: %s" % repr(ack))
        raise errors.RadioError("Radio NAK'd block at %04x" % addr)
    _cmd, _addr, _length = struct.unpack(">cHb", header)
    if _addr != addr or _length != _length:
        LOG.debug("Expected/Received:")
        LOG.debug(" Length: %02x/%02x" % (length, _length))
        LOG.debug(" Addr: %04x/%04x" % (addr, _addr))
        raise errors.RadioError("Radio send unexpected block")
    cs = _checksum(result[1:-2])
    if cs != result[-2]:
        LOG.debug("Calculated: %02x" % cs)
        LOG.debug("Actual:     %02x" % result[-2])
        raise errors.RadioError("Block at 0x%04x failed checksum" % addr)
    return data


def _finish(radio):
    endframe = b"\x45\x4E\x44"
    _echo_write(radio, endframe)
    result = radio.pipe.read(1)
    # TYT radios acknowledge the "endframe" command, Luiton radios do not.
    if result != b"" and result != b"\x06":
        LOG.error("Got:\n%s" % util.hexprint(result))
        raise errors.RadioError("Radio did not finish cleanly")


def do_download(radio):

    _ident(radio)

    _memobj = None
    data = b""

    for start, end in radio._ranges:
        for addr in range(start, end, 0x10):
            block = _send(radio, b'R', addr, 0x10)
            data += block
            status = chirp_common.Status()
            status.cur = len(data)
            status.max = end
            status.msg = "Downloading from radio"
            radio.status_fn(status)

    _finish(radio)

    return memmap.MemoryMapBytes(data)


def do_upload(radio):

    _ident(radio)

    for start, end in radio._ranges:
        for addr in range(start, end, 0x10):
            if addr < 0x0100:
                continue
            block = radio._mmap[addr:addr + 0x10]
            _send(radio, b'W', addr, len(block), block)
            status = chirp_common.Status()
            status.cur = addr
            status.max = end
            status.msg = "Uploading to Radio"
            radio.status_fn(status)

    _finish(radio)


#
# The base class, extended for use with other models
#
class Th9000Radio(chirp_common.CloneModeRadio,
                  chirp_common.ExperimentalRadio):
    """TYT TH-9000"""
    VENDOR = "TYT"
    MODEL = "TH9000 Base"
    BAUD_RATE = 9600
    NEEDS_COMPAT_SERIAL = False
    valid_freq = [(900000000, 999000000)]

    _memsize = MMAPSIZE
    _ranges = [(0x0000, 0x4000)]

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = ("The TYT TH-9000 driver is an beta version."
                           "Proceed with Caution and backup your data")
        return rp

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_bank = False
        rf.has_cross = True
        rf.has_tuning_step = False
        rf.has_rx_dtcs = True
        rf.valid_skips = ["", "S"]
        rf.memory_bounds = (0, 199)
        rf.valid_name_length = 7
        rf.valid_characters = chirp_common.CHARSET_UPPER_NUMERIC + "-"
        rf.valid_modes = MODES
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'DTCS', 'Cross']
        rf.valid_cross_modes = ['Tone->DTCS', 'DTCS->Tone',
                                '->Tone', '->DTCS', 'Tone->Tone']
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_tones = TONES
        rf.valid_dtcs_codes = chirp_common.ALL_DTCS_CODES
        rf.valid_bands = self.valid_freq
        rf.valid_tuning_steps = TUNING_STEPS
        return rf

    # Do a download of the radio from the serial port
    def sync_in(self):
        self._mmap = do_download(self)
        self.process_mmap()

    # Do an upload of the radio to the serial port
    def sync_out(self):
        do_upload(self)

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    # Return a raw representation of the memory object, which
    # is very helpful for development
    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number])

    # not working yet
    def _get_dcs_index(self, _mem, which):
        base = getattr(_mem, '%scode' % which)
        extra = getattr(_mem, '%sdcsextra' % which)
        return (int(extra) << 8) | int(base)

    def _set_dcs_index(self, _mem, which, index):
        base = getattr(_mem, '%scode' % which)
        extra = getattr(_mem, '%sdcsextra' % which)
        base.set_value(index & 0xFF)
        extra.set_value(index >> 8)

    # Extract a high-level memory object from the low-level memory map
    # This is called to populate a memory in the UI
    def get_memory(self, number):
        # Get a low-level memory object mapped to the image
        _mem = self._memobj.memory[number]

        # get flag info
        cbyte = number // 8
        cbit = 7 - (number % 8)
        setflag = self._memobj.csetflag[cbyte].c[cbit]
        skipflag = self._memobj.cskipflag[cbyte].c[cbit]

        mem = chirp_common.Memory()

        mem.number = number  # Set the memory number

        if setflag == 1:
            mem.empty = True
            return mem

        mem.freq = int(_mem.freq) * 100

        # compensate for 6.25 and 12.5 kHz tuning steps, add 500 Hz if needed
        lastdigit = int(_mem.freq) % 10
        if (lastdigit == 2 or lastdigit == 7):
            mem.freq += 50

        mem.offset = int(_mem.offset) * 100
        # Set the alpha tag
        mem.name = self.filter_name(str(_mem.name).rstrip())
        mem.duplex = DUPLEXES[_mem.duplex]
        mem.mode = MODES[_mem.channel_width]
        mem.power = POWER_LEVELS[_mem.power]

        rxtone = txtone = None

        rxmode = TMODES[_mem.rxtmode]
        txmode = TMODES[_mem.txtmode]

        # doesn't work
        if rxmode == "Tone":
            rxtone = TONES[_mem.rxtone]
        elif rxmode == "DTCS":
            rxtone = chirp_common.ALL_DTCS_CODES[self._get_dcs_index(
                                                 _mem, 'rx')]

        if txmode == "Tone":
            txtone = TONES[_mem.txtone]
        elif txmode == "DTCS":
            txtone = chirp_common.ALL_DTCS_CODES[self._get_dcs_index(
                                                 _mem, 'tx')]

        rxpol = _mem.rxinv and "R" or "N"
        txpol = _mem.txinv and "R" or "N"

        chirp_common.split_tone_decode(mem,
                                       (txmode, txtone, txpol),
                                       (rxmode, rxtone, rxpol))

        mem.skip = "S" if skipflag == 1 else ""

        # We'll consider any blank (i.e. 0 MHz frequency) to be empty
        if mem.freq == 0:
            mem.empty = True

        return mem

    # Store details about a high-level memory to the memory map
    # This is called when a user edits a memory in the UI
    def set_memory(self, mem):
        # Get a low-level memory object mapped to the image

        _mem = self._memobj.memory[mem.number]

        cbyte = mem.number // 8
        cbit = 7 - (mem.number % 8)

        if mem.empty:
            self._memobj.csetflag[cbyte].c[cbit] = 1
            self._memobj.cskipflag[cbyte].c[cbit] = 1
            return

        self._memobj.csetflag[cbyte].c[cbit] = 0
        self._memobj.cskipflag[cbyte].c[cbit] = 1 if (mem.skip == "S") else 0

        _mem.set_raw("\x00" * 32)

        _mem.freq = mem.freq / 100         # Convert to low-level frequency
        _mem.offset = mem.offset / 100         # Convert to low-level frequency

        _mem.name = mem.name.ljust(7)[:7]  # Store the alpha tag
        _mem.duplex = DUPLEXES.index(mem.duplex)

        try:
            _mem.channel_width = MODES.index(mem.mode)
        except ValueError:
            _mem.channel_width = 0

        ((txmode, txtone, txpol),
         (rxmode, rxtone, rxpol)) = chirp_common.split_tone_encode(mem)

        _mem.txtmode = TMODES.index(txmode)

        _mem.rxtmode = TMODES.index(rxmode)

        if txmode == "Tone":
            _mem.txtone = TONES.index(txtone)
        elif txmode == "DTCS":
            self._set_dcs_index(_mem, 'tx',
                                chirp_common.ALL_DTCS_CODES.index(txtone))

        if rxmode == "Tone":
            _mem.rxtone = TONES.index(rxtone)
        elif rxmode == "DTCS":
            self._set_dcs_index(_mem, 'rx',
                                chirp_common.ALL_DTCS_CODES.index(rxtone))

        _mem.txinv = txpol == "R"
        _mem.rxinv = rxpol == "R"

        if mem.power:
            _mem.power = POWER_LEVELS.index(mem.power)
        else:
            _mem.power = 0

    def _get_settings(self):
        _settings = self._memobj.settings
        _freqrange = self._memobj.freqrange
        _slabel = self._memobj.slabel

        basic = RadioSettingGroup("basic", "Global Settings")
        freqrange = RadioSettingGroup("freqrange", "Frequency Ranges")
        top = RadioSettingGroup("top", "All Settings", basic, freqrange)
        settings = RadioSettings(top)

        def _filter(name):
            filtered = ""
            for char in str(name):
                if char in chirp_common.CHARSET_ASCII:
                    filtered += char
                else:
                    filtered += ""
            return filtered

        val = RadioSettingValueString(0, 7, _filter(_slabel.startname))
        rs = RadioSetting("startname", "Startup Label", val)
        basic.append(rs)

        rs = RadioSetting("bg_color", "LCD Color",
                          RadioSettingValueList(BGCOLOR_LIST, BGCOLOR_LIST[
                                                _settings.bg_color]))
        basic.append(rs)

        rs = RadioSetting("bg_brightness", "LCD Brightness",
                          RadioSettingValueList(BGBRIGHT_LIST, BGBRIGHT_LIST[
                                                _settings.bg_brightness]))
        basic.append(rs)

        rs = RadioSetting("squelch", "Squelch Level",
                          RadioSettingValueList(SQUELCH_LIST, SQUELCH_LIST[
                                                _settings.squelch]))
        basic.append(rs)

        rs = RadioSetting("timeout_timer", "Timeout Timer (TOT)",
                          RadioSettingValueList(TIMEOUT_LIST, TIMEOUT_LIST[
                                                _settings.timeout_timer]))
        basic.append(rs)

        rs = RadioSetting("auto_power_off", "Auto Power Off (APO)",
                          RadioSettingValueList(APO_LIST, APO_LIST[
                                                _settings.auto_power_off]))
        basic.append(rs)

        rs = RadioSetting("voice_prompt", "Beep Prompt",
                          RadioSettingValueList(BEEP_LIST, BEEP_LIST[
                                                 _settings.voice_prompt]))
        basic.append(rs)

        rs = RadioSetting("tbst_freq", "Tone Burst Frequency",
                          RadioSettingValueList(TBSTFREQ_LIST, TBSTFREQ_LIST[
                                                _settings.tbst_freq]))
        basic.append(rs)

        rs = RadioSetting("choose_tx_power", "Max Level of TX Power",
                          RadioSettingValueList(TXPWR_LIST, TXPWR_LIST[
                                                _settings.choose_tx_power]))
        basic.append(rs)

        (flow, fhigh) = self.valid_freq[0]
        flow /= 1000
        fhigh /= 1000
        fmidrange = (fhigh - flow) / 2

        rs = RadioSetting("txrangelow", "TX Freq, Lower Limit (kHz)",
                          RadioSettingValueInteger(
                              flow, flow + fmidrange,
                              int(_freqrange.txrangelow) / 10))
        freqrange.append(rs)

        rs = RadioSetting("txrangehi", "TX Freq, Upper Limit (kHz)",
                          RadioSettingValueInteger(
                              fhigh-fmidrange, fhigh,
                              int(_freqrange.txrangehi) / 10))
        freqrange.append(rs)

        rs = RadioSetting("rxrangelow", "RX Freq, Lower Limit (kHz)",
                          RadioSettingValueInteger(
                              flow, flow+fmidrange,
                              int(_freqrange.rxrangelow) / 10))
        freqrange.append(rs)

        rs = RadioSetting("rxrangehi", "RX Freq, Upper Limit (kHz)",
                          RadioSettingValueInteger(
                              fhigh-fmidrange, fhigh,
                              int(_freqrange.rxrangehi) / 10))
        freqrange.append(rs)

        return settings

    def get_settings(self):
        try:
            return self._get_settings()
        except:
            import traceback
            LOG.error("failed to parse settings")
            traceback.print_exc()
            return None

    def set_settings(self, settings):
        _settings = self._memobj.settings
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    name = element.get_name()

                    if name in ["txrangelow", "txrangehi", "rxrangelow",
                                "rxrangehi"]:
                        LOG.debug("setting %s = %s" % (name,
                                  int(element.value) * 10))
                        setattr(self._memobj.freqrange, name,
                                int(element.value) * 10)
                        continue

                    if name in ["startname"]:
                        LOG.debug("setting %s = %s" % (name, element.value))
                        setattr(self._memobj.slabel, name, element.value)
                        continue

                    obj = _settings
                    setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("using apply callback")
                        element.run_apply_callback()
                    else:
                        LOG.debug("Setting %s = %s" % (setting,
                                  element.value))
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise

    @classmethod
    def match_model(cls, filedata, filename):
        return False


def match_orig_model(cls, filedata, filename):
    # This old-style file detection should only be used for the
    # original TYT TH9000 classes for compatibility
    if cls.VENDOR != 'TYT' or 'TH9000_' not in cls.MODEL:
        return False

    if MMAPSIZE == len(filedata):
        (flow, fhigh) = cls.valid_freq[0]
        flow /= 1000000
        fhigh /= 1000000

        txmin = filedata[0x200] * 100 + (filedata[0x201] >> 4) \
            * 10 + filedata[0x201] % 16
        txmax = filedata[0x204] * 100 + (filedata[0x205] >> 4) \
            * 10 + filedata[0x205] % 16
        rxmin = filedata[0x208] * 100 + (filedata[0x209] >> 4) \
            * 10 + filedata[0x209] % 16
        rxmax = filedata[0x20C] * 100 + (filedata[0x20D] >> 4) \
            * 10 + filedata[0x20D] % 16

        if (rxmin >= flow and rxmax <= fhigh and txmin >= flow and
                txmax <= fhigh):
            return True

    return False


@directory.register
class Th9000220Radio(Th9000Radio):
    """TYT TH-9000 220"""
    VENDOR = "TYT"
    MODEL = "TH9000_220"
    BAUD_RATE = 9600
    valid_freq = [(220000000, 260000000)]

    @classmethod
    def match_model(cls, filedata, filename):
        return match_orig_model(cls, filedata, filename)


@directory.register
class Th9000144Radio(Th9000220Radio):
    """TYT TH-9000 144"""
    VENDOR = "TYT"
    MODEL = "TH9000_144"
    BAUD_RATE = 9600
    valid_freq = [(136000000, 174000000)]

    @classmethod
    def match_model(cls, filedata, filename):
        return match_orig_model(cls, filedata, filename)


@directory.register
class Th9000440Radio(Th9000220Radio):
    """TYT TH-9000 440"""
    VENDOR = "TYT"
    MODEL = "TH9000_440"
    BAUD_RATE = 9600
    valid_freq = [(400000000, 490000000)]

    @classmethod
    def match_model(cls, filedata, filename):
        return match_orig_model(cls, filedata, filename)


@directory.register
class Lt580VHFRadio(Th9000144Radio):
    """Luiton LT-580 VHF"""
    VENDOR = "LUITON"
    MODEL = "LT-580_VHF"


@directory.register
class Lt580UHFRadio(Th9000440Radio):
    """Luiton LT-580 UHF"""
    VENDOR = "LUITON"
    MODEL = "LT-580_UHF"


@directory.register
class RT9000DVHFRadio(Th9000Radio):
    """Retevis RT9000D VHF"""
    VENDOR = "Retevis"
    MODEL = "RT9000D_136-174"
    valid_freq = [(136000000, 174000000)]


@directory.register
class RT9000D220Radio(Th9000Radio):
    """Retevis RT9000D 220"""
    VENDOR = "Retevis"
    MODEL = "RT9000D_220-260"
    valid_freq = [(220000000, 260000000)]


@directory.register
class RT9000DUHFRadio(Th9000Radio):
    """Retevis RT9000D UHF"""
    VENDOR = "Retevis"
    MODEL = "RT9000D_400-490"
    valid_freq = [(400000000, 490000000)]


@directory.register
class RT9000D6688Radio(Th9000Radio):
    """Retevis RT9000D 66-88"""
    VENDOR = "Retevis"
    MODEL = "RT9000D_66-88"
    valid_freq = [(66000000, 88000000)]
